from functools import partial
import pickle

import pytest
from hypothesis import given, strategies
import hypothesis.extra.numpy as npst
from packaging import version

from scipy import special
from scipy.special._support_alternative_backends import _special_funcs
from scipy._lib._array_api_no_0d import xp_assert_close
from scipy._lib._array_api import (is_cupy, is_dask, is_jax, is_torch,
                                   make_xp_pytest_param, make_xp_test_case,
                                   get_native_namespace_name)
from scipy._external.array_api_compat import numpy as np
import scipy._external.array_api_extra as xpx


# Run all tests in this module in the Array API CI, including those without
# the xp fixture
pytestmark = pytest.mark.array_api_backends

lazy_xp_modules = [special]


def _skip_or_tweak_alternative_backends(xp, nfo, dtypes, int_only):
    """Skip tests for specific intersections of scipy.special functions 
    vs. backends vs. dtypes vs. devices.
    Also suggest bespoke tweaks.

    Returns
    -------
    positive_only : list[bool]
        Whether you should exclusively test positive inputs.
    dtypes : list[str]
        dtype strings 'float64', 'int32', 'int64', etc. with integer types
        mapped to the type of the NumPy default int.
    """
    f_name = nfo.name
    if isinstance(nfo.positive_only, dict):
        positive_only = nfo.positive_only.get(get_native_namespace_name(xp), False)
    else:
        positive_only = nfo.positive_only
    if isinstance(positive_only, bool):
        positive_only = [positive_only]*nfo.n_args

    dtypes = [np.intp.__name__ if dtype == "intp" else dtype for dtype in dtypes]

    if f_name in {'betaincinv'} and is_cupy(xp):
        pytest.xfail("CuPy uses different convention for out of domain input.")

    if not any('int' in dtype for dtype in dtypes):
        return positive_only, dtypes

    # Integer-specific issues from this point onwards

    if f_name in {'gamma', 'gammasgn'} and is_cupy(xp):
        # CuPy has not yet updated gamma pole behavior to match
        # https://github.com/scipy/scipy/pull/21827.
        positive_only = [True]

    if f_name in {'poch'} and is_jax(xp):
        # Jax uses a different convention at gamma poles.
        positive_only = [True, True]

    if f_name == 'multigammaln':
        pytest.skip("multigammaln raises for out of domain inputs.")

    if ((is_torch(xp) and f_name in {'gammainc', 'gammaincc'})
        or (is_cupy(xp) and f_name in {'stdtr', 'i0e', 'i1e'})
        or (is_jax(xp) and f_name in {'stdtr', 'ndtr', 'ndtri', 'log_ndtr', 'hyp1f1',
                                      'hyp2f1', 'spence', 'kl_div'})
    ):
        pytest.skip(f"`{f_name}` does not support integer types")

    # int/float mismatched args support is sketchy
    if (any('float' in dtype for dtype in dtypes)
        and ((is_torch(xp) and f_name in ('rel_entr', 'xlogy', 'polygamma',
                                          'zeta', 'xlog1py'))
             or (is_jax(xp) and f_name in ('gammainc', 'gammaincc', 'expn',
                                           'rel_entr', 'xlogy', 'betaln',
                                           'polygamma', 'zeta', 'poch',
                                           'xlog1py')))
    ):
        pytest.xfail("dtypes do not match")

    if (is_torch(xp) and xpx.default_dtype(xp) == xp.float32):
        # On PyTorch with float32 default dtype, all ints are promoted
        # to float32, but when falling back to NumPy/SciPy int64 is promoted
        # instead to float64. Integer only parameters essentially do not
        # participate in determination of the result type in PyTorch with
        # float32 default dtype, but will impact the output dtype as if
        # they were float64 when falling back to NumPy/SciPy.
        if not nfo.torch_native:
            pytest.xfail("dtypes do not match")

    return positive_only, dtypes


@pytest.mark.filterwarnings("ignore:invalid value encountered:RuntimeWarning:dask")
@pytest.mark.filterwarnings("ignore:overflow encountered:RuntimeWarning")
@pytest.mark.parametrize('shapes', [[(0,)]*4, [tuple()]*4, [(10,)]*4,
                                    [(10,), (11, 1), (12, 1, 1), (13, 1, 1, 1)]])
@pytest.mark.parametrize('base_dtype', ['float32', 'float64', 'intp'])
@pytest.mark.parametrize(
    'func,nfo', [make_xp_pytest_param(i.wrapper, i) for i in _special_funcs])
def test_support_alternative_backends(xp, func, nfo, base_dtype, shapes):
    int_only = nfo.int_only
    if int_only is None:
        int_only = (False, ) * nfo.n_args
        dtypes = (base_dtype, ) * nfo.n_args
    else:
        dtypes = tuple(
            'intp' if needs_int else base_dtype for needs_int in int_only
        )

    positive_only, dtypes = _skip_or_tweak_alternative_backends(
        xp, nfo, dtypes, int_only
    )

    dtypes_np = [getattr(np, dtype) for dtype in dtypes]
    dtypes_xp = [getattr(xp, dtype) for dtype in dtypes]

    shapes = shapes[:nfo.n_args]
    rng = np.random.default_rng(984254252920492019)
    args_np = []

    # Handle cases where there's an argument which only takes scalar values.
    python_int_only = nfo.python_int_only
    if isinstance(python_int_only, dict):
        python_int_only = python_int_only.get(get_native_namespace_name(xp))
    scalar_or_0d_only = nfo.scalar_or_0d_only
    if isinstance(scalar_or_0d_only, dict):
        scalar_or_0d_only = scalar_or_0d_only.get(get_native_namespace_name(xp))

    test_large_ints = nfo.test_large_ints
    if isinstance(nfo.test_large_ints, dict):
        test_large_ints = test_large_ints.get(get_native_namespace_name(xp), False)

    if python_int_only is None:
        python_int_only = [False] * nfo.n_args
    if scalar_or_0d_only is None:
        scalar_or_0d_only = [False] * nfo.n_args

    no_shape = [
        cond1 or cond2 for cond1, cond2 in zip(python_int_only, scalar_or_0d_only)
    ]

    shapes = [shape if not cond else None for shape, cond in zip(shapes, no_shape)]

    for dtype, dtype_np, shape, needs_python_int in zip(
            dtypes, dtypes_np, shapes, python_int_only
    ):
        if 'int' in dtype and test_large_ints:
            iinfo = np.iinfo(dtype_np)
            rand = partial(rng.integers, iinfo.min, iinfo.max + 1)
        elif 'int' in dtype:
            rand = partial(rng.integers, -20, 21)
        else:
            rand = rng.standard_normal
        val = rand(size=shape, dtype=dtype_np)
        if needs_python_int:
            # The logic above for determining shapes guarantees that
            # shape will be None in the above line when a Python int is required,
            # so this can safely be converted to an int.
            val = int(val)
        args_np.append(val)

    args_np = [
        np.abs(arg) if cond else arg for arg, cond in zip(args_np, positive_only)
    ]

    args_xp = [
        xp.asarray(arg, dtype=dtype_xp) if not needs_python_int
        else arg
        for arg, dtype_xp, needs_python_int
        in zip(args_np, dtypes_xp, python_int_only)
    ]

    args_np = [
        np.asarray(arg, dtype=dtype_np) if not needs_python_int
        else arg
        for arg, dtype_np, needs_python_int
        in zip(args_np, dtypes_np, python_int_only)
    ]

    if is_dask(xp):
        # We're using map_blocks to dispatch the function to Dask.
        # This is the correct thing to do IF all tested functions are elementwise;
        # otherwise the output would change depending on chunking.
        # Try to trigger bugs related to having multiple chunks.
        args_xp = [arg.rechunk(5) for arg in args_xp]

    res = nfo.wrapper(*args_xp)  # Also wrapped by lazy_xp_function
    ref = nfo.func(*args_np)  # Unwrapped ufunc
    if (
            is_torch(xp)
            and xpx.default_dtype(xp) == xp.float32
            and "float64" not in dtypes
    ):
        # int64 promotes like float32 on torch with default dtype = float32
        # cast reference if needed
        ref = np.float32(ref)
    # When dtype_np is integer, the output dtype can be float
    atol = 0 if ref.dtype.kind in 'iu' else 10 * np.finfo(ref.dtype).eps
    rtol = None
    if is_torch(xp) and func.__name__ == 'j1':
        # If we end up needing more function/backend specific tolerance
        # adjustments, this should be factored out properly.
        atol = 1e-7
        rtol = 1e-5
    xp_assert_close(
        res, xp.asarray(ref), rtol=rtol, atol=atol, check_0d=nfo.produces_0d
    )


@pytest.mark.parametrize(
    'func, nfo',
    [make_xp_pytest_param(i.wrapper, i) for i in _special_funcs if i.n_args >= 2])
@pytest.mark.filterwarnings("ignore:invalid value encountered:RuntimeWarning:dask")
def test_support_alternative_backends_mismatched_dtypes(xp, func, nfo):
    """Test mix-n-match of int and float arguments"""
    if func.__name__ in {'expn', 'polygamma', 'multigammaln', 'bdtr', 'bdtrc', 'bdtri',
                         'nbdtr', 'nbdtrc', 'nbdtri', 'pdtri'}:
        pytest.skip(f"dtypes for {func.__name__} make it a bad fit for this test.")
    dtypes = ['intp', 'float32', 'float64', 'float64'][:nfo.n_args]

    positive_only, dtypes = _skip_or_tweak_alternative_backends(
        xp, nfo, dtypes, (False,)*nfo.n_args
    )
    dtypes_np = [getattr(np, dtype) for dtype in dtypes]
    dtypes_xp = [getattr(xp, dtype) for dtype in dtypes]

    rng = np.random.default_rng(984254252920492019)
    iinfo = np.iinfo(np.intp)
    if nfo.test_large_ints:
        randint = partial(rng.integers, iinfo.min, iinfo.max + 1)
    else:
        randint = partial(rng.integers, -20, 21)
    args_np = [
        randint(size=1, dtype=np.int64),
        rng.standard_normal(size=1, dtype=np.float32),
        rng.standard_normal(size=1, dtype=np.float64),
        rng.standard_normal(size=1, dtype=np.float64),
    ][:nfo.n_args]
    args_np = [
        np.abs(arg) if cond else arg for arg, cond in zip(args_np, positive_only)
    ]

    args_xp = [xp.asarray(arg, dtype=dtype_xp)
               for arg, dtype_xp in zip(args_np, dtypes_xp)]
    args_np = [np.asarray(arg, dtype=dtype_np)
               for arg, dtype_np in zip(args_np, dtypes_np)]

    res = nfo.wrapper(*args_xp)  # Also wrapped by lazy_xp_function
    ref = nfo.func(*args_np)  # Unwrapped ufunc
    if (
            is_torch(xp)
            and xpx.default_dtype(xp) == xp.float32
            and "float64" not in dtypes
    ):
        # int64 promotes like float32 on torch with default dtype = float32
        # cast reference if needed
        ref = np.float32(ref)

    atol = 10 * np.finfo(ref.dtype).eps
    xp_assert_close(res, xp.asarray(ref), atol=atol)


@pytest.mark.xslow
@given(data=strategies.data())
@pytest.mark.fail_slow(5)
@pytest.mark.parametrize(
    'func,nfo', [make_xp_pytest_param(nfo.wrapper, nfo) for nfo in _special_funcs])
@pytest.mark.filterwarnings("ignore:invalid value encountered:RuntimeWarning:dask")
@pytest.mark.filterwarnings("ignore:divide by zero encountered:RuntimeWarning:dask")
@pytest.mark.filterwarnings("ignore:overflow encountered:RuntimeWarning:dask")
@pytest.mark.filterwarnings(
    "ignore:overflow encountered:RuntimeWarning:array_api_strict"
)
def test_support_alternative_backends_hypothesis(xp, func, nfo, data):
    if func.__name__ in {'expn', 'polygamma', 'multigammaln', 'bdtr', 'bdtrc', 'bdtri',
                         'nbdtr', 'nbdtrc', 'nbdtri', 'pdtri'}:
        pytest.skip(f"dtypes for {func.__name__} make it a bad fit for this test.")
    dtype = data.draw(strategies.sampled_from(['float32', 'float64', 'intp']))
    positive_only, dtypes = _skip_or_tweak_alternative_backends(
        xp, nfo, [dtype], (False,)*nfo.n_args
    )
    dtype_np = getattr(np, dtypes[0])
    dtype_xp = getattr(xp, dtypes[0])

    elements = {'allow_subnormal': False}
    # Most failures are due to NaN or infinity; uncomment to suppress them
    # elements['allow_infinity'] = False
    # elements['allow_nan'] = False
    if any(positive_only):
        elements['min_value'] = 0

    shapes, _ = data.draw(
        npst.mutually_broadcastable_shapes(num_shapes=nfo.n_args))
    args_np = [data.draw(npst.arrays(dtype_np, shape, elements=elements))
               for shape in shapes]

    args_xp = [xp.asarray(arg, dtype=dtype_xp) for arg in args_np]
    args_np = [np.asarray(arg, dtype=dtype_np) for arg in args_np]

    res = nfo.wrapper(*args_xp)  # Also wrapped by lazy_xp_function
    ref = nfo.func(*args_np)  # Unwrapped ufunc
    if (
            is_torch(xp)
            and xpx.default_dtype(xp) == xp.float32
            and dtype != "float64"
    ):
        # int64 promotes like float32 on torch with default dtype = float32
        # cast reference if needed
        ref = np.float32(ref)

    # When dtype_np is integer, the output dtype can be float
    atol = 0 if ref.dtype.kind in 'iu' else 10 * np.finfo(ref.dtype).eps
    xp_assert_close(res, xp.asarray(ref), atol=atol)


@pytest.mark.filterwarnings("ignore:numpy.core is deprecated:DeprecationWarning")
@pytest.mark.parametrize("func", [nfo.wrapper for nfo in _special_funcs])
def test_pickle(func):
    roundtrip = pickle.loads(pickle.dumps(func))
    assert roundtrip is func


@pytest.mark.parametrize("func", [nfo.wrapper for nfo in _special_funcs])
def test_repr(func):
    assert func.__name__ in repr(func)
    assert "locals" not in repr(func)


@pytest.mark.skipif(
    version.parse(np.__version__) < version.parse("2.2"),  # type:ignore[attr-defined]
    reason="Can't update ufunc __doc__ when SciPy is compiled vs. NumPy < 2.2")
@pytest.mark.parametrize('func', [nfo.wrapper for nfo in _special_funcs])
def test_doc(func):
    """xp_capabilities updates the docstring in place. 
    Make sure it does so exactly once, including when SCIPY_ARRAY_API is not set.
    """
    match = "has experimental support for Python Array API"
    assert func.__doc__.count(match) == 1


@pytest.mark.parametrize(
    'func,n_args,int_only,is_ufunc',
    [(nfo.wrapper, nfo.n_args, nfo.int_only, nfo.is_ufunc)
     for nfo in _special_funcs]
)
def test_ufunc_kwargs(func, n_args, int_only, is_ufunc):
    """Test that numpy-specific out= and dtype= keyword arguments
    of ufuncs still work when SCIPY_ARRAY_API is set.
    """
    if not is_ufunc:
        pytest.skip(f"{func.__name__} is not a ufunc.")
    if int_only is None:
        int_only = (False, ) * n_args
    # out=
    args = [
        np.asarray([.1, .2]) if not needs_int
        else np.asarray([1, 2])
        for needs_int in int_only
    ]
    out = np.empty(2)
    y = func(*args, out=out)
    xp_assert_close(y, out)

    # out= with out.dtype != args.dtype
    out = np.empty(2, dtype=np.float32)
    y = func(*args, out=out)
    xp_assert_close(y, out)

    if func.__name__ in {"bdtr", "bdtrc", "bdtri"}:
        # The below function evaluation will trigger a deprecation warning
        # with dtype=np.float32. This will go away if the trigger is actually
        # pulled on the deprecation.
        return

    # dtype=
    y = func(*args, dtype=np.float32)
    assert y.dtype == np.float32


@make_xp_test_case(special.chdtr)
def test_chdtr_gh21311(xp):
    # the edge case behavior of generic chdtr was not right; see gh-21311
    # be sure to test at least these cases
    # should add `np.nan` into the mix when gh-21317 is resolved
    x = np.asarray([-np.inf, -1., 0., 1., np.inf])
    v = x.reshape(-1, 1)
    ref = special.chdtr(v, x)
    res = special.chdtr(xp.asarray(v), xp.asarray(x))
    xp_assert_close(res, xp.asarray(ref))


@make_xp_test_case(special.fdtrc)
def test_mixed_arrays_and_python_scalars(xp):
    # Tests that the delegation infrastructure respects NEP50.
    res = special.fdtrc(1.1, 2., xp.asarray(1., dtype=xp.float32))
    ref = xp.asarray(0.4349004, dtype=xp.float32)
    xp_assert_close(res, ref)



"""DEPRECATED - This module is kept here only as a backward compatibility shim
for the old ufoLib.etree module, which was moved to :mod:`fontTools.misc.etree`.
Please use the latter instead.
"""

from fontTools.misc.etree import *



"""IFT - Incremental Font Transfer table."""

# Reference: https://github.com/googlefonts/fontations/blob/main/read-fonts/src/tables/ift.rs


from .otBase import BaseTTXConverter


class table_I_F_T_X_(BaseTTXConverter):
    """Incremental Font Transfer table."""

    pass



import numpy as np


EPS = np.finfo(float).eps


def get_arrays_tol(*arrays):
    """
    Get a relative tolerance for a set of arrays.

    Parameters
    ----------
    *arrays: tuple
        Set of `numpy.ndarray` to get the tolerance for.

    Returns
    -------
    float
        Relative tolerance for the set of arrays.

    Raises
    ------
    ValueError
        If no array is provided.
    """
    if len(arrays) == 0:
        raise ValueError("At least one array must be provided.")
    size = max(array.size for array in arrays)
    weight = max(
        np.max(np.abs(array[np.isfinite(array)]), initial=1.0)
        for array in arrays
    )
    return 10.0 * EPS * max(size, 1.0) * weight


def exact_1d_array(x, message):
    """
    Preprocess a 1-dimensional array.

    Parameters
    ----------
    x : array_like
        Array to be preprocessed.
    message : str
        Error message if `x` cannot be interpreter as a 1-dimensional array.

    Returns
    -------
    `numpy.ndarray`
        Preprocessed array.
    """
    x = np.atleast_1d(np.squeeze(x)).astype(float)
    if x.ndim != 1:
        raise ValueError(message)
    return x


def exact_2d_array(x, message):
    """
    Preprocess a 2-dimensional array.

    Parameters
    ----------
    x : array_like
        Array to be preprocessed.
    message : str
        Error message if `x` cannot be interpreter as a 2-dimensional array.

    Returns
    -------
    `numpy.ndarray`
        Preprocessed array.
    """
    x = np.atleast_2d(x).astype(float)
    if x.ndim != 2:
        raise ValueError(message)
    return x



"""This module contains miscellaneous helpers.

It is not considered part of the public ufoLib API. It does, however,
define the :py:obj:`.deprecated` decorator that is used elsewhere in
the module.
"""

from __future__ import annotations

from typing import Optional, Type, TypeVar, Union, cast
from collections.abc import Callable
import enum
import functools
import warnings

F = TypeVar("F", bound=Callable[..., object])
FormatVersion = TypeVar("FormatVersion", bound="BaseFormatVersion")
FormatVersionInput = Optional[Union[int, tuple[int, int], FormatVersion]]

numberTypes = (int, float)


def deprecated(msg: str = "") -> Callable[[F], F]:
    """Decorator factory to mark functions as deprecated with given message.

    >>> @deprecated("Enough!")
    ... def some_function():
    ...    "I just print 'hello world'."
    ...    print("hello world")
    >>> some_function()
    hello world
    >>> some_function.__doc__ == "I just print 'hello world'."
    True
    """

    def deprecated_decorator(func: F) -> F:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(
                f"{func.__name__} function is a deprecated. {msg}",
                category=DeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return cast(F, wrapper)

    return deprecated_decorator


def normalizeFormatVersion(
    value: FormatVersionInput, cls: Type[FormatVersion]
) -> FormatVersion:
    # Needed for type safety of UFOFormatVersion and GLIFFormatVersion input
    if value is None:
        return cls.default()
    if isinstance(value, cls):
        return value
    if isinstance(value, int):
        return cls((value, 0))
    if isinstance(value, tuple) and len(value) == 2:
        return cls(value)
    raise ValueError(f"Unsupported format version: {value!r}")


# Base class for UFOFormatVersion and GLIFFormatVersion
class BaseFormatVersion(tuple[int, int], enum.Enum):
    value: tuple[int, int]

    def __new__(cls: Type[FormatVersion], value: tuple[int, int]) -> BaseFormatVersion:
        return super().__new__(cls, value)

    @property
    def major(self) -> int:
        return self.value[0]

    @property
    def minor(self) -> int:
        return self.value[1]

    @classmethod
    def _missing_(cls, value: object) -> BaseFormatVersion:
        # allow to initialize a version enum from a single (major) integer
        if isinstance(value, int):
            return cls((value, 0))
        # or from None to obtain the current default version
        if value is None:
            return cls.default()
        raise ValueError(f"{value!r} is not a valid {cls.__name__}")

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}"

    @classmethod
    def default(cls: Type[FormatVersion]) -> FormatVersion:
        # get the latest defined version (i.e. the max of all versions)
        return max(cls.__members__.values())

    @classmethod
    def supported_versions(cls: Type[FormatVersion]) -> frozenset[FormatVersion]:
        return frozenset(cls.__members__.values())


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""SVD decomposition functions."""
import numpy as np

from scipy._lib._util import _apply_over_batch, _deprecate_dtypes
from . import _batched_linalg

# Local imports.
from ._misc import LinAlgError, _datacopied
from .lapack import _normalize_lapack_dtype, _ensure_aligned_and_native, HAS_ILP64
from scipy.linalg.lapack import get_lapack_funcs   # noqa: F401  (backwards compat)
from ._decomp import _asarray_validated


__all__ = ['svd', 'svdvals', 'diagsvd', 'orth', 'subspace_angles', 'null_space']


def _format_emit_errors_warnings(err_lst, lapack_driver):
    """Format/emit errors/warnings from a lowlevel batched routine.
    """
    # NB the low-level routine currently stops processing a batch at the first error
    for entry in err_lst:
        info = entry["lapack_info"]
        num = entry["num"]
        if info != 0:
            if info > 0:
                raise LinAlgError(f"SVD did not converge for slice = {num}.")
            if info < 0:
                if lapack_driver == "gesdd" and info == -4:
                    msg = f"slice {num} has a NaN entry"
                    raise ValueError(msg)
                raise ValueError(
                    f'illegal value in {-info}th argument of internal {lapack_driver}'
                    f'  for slice {num}.'
                )


def svd(a, full_matrices=True, compute_uv=True, overwrite_a=False,
        check_finite=True, lapack_driver='gesdd'):
    """
    Singular Value Decomposition.

    Factorizes the matrix `a` into two unitary matrices ``U`` and ``Vh``, and
    a 1-D array ``s`` of singular values (real, non-negative) such that
    ``a == U @ S @ Vh``, where ``S`` is a suitably shaped matrix of zeros with
    main diagonal ``s``.

    Parameters
    ----------
    a : (..., M, N) array_like
        Matrix to decompose.
    full_matrices : bool, optional
        If True (default), `U` and `Vh` are of shape ``(M, M)``, ``(N, N)``.
        If False, the shapes are ``(M, K)`` and ``(K, N)``, where
        ``K = min(M, N)``.
    compute_uv : bool, optional
        Whether to compute also ``U`` and ``Vh`` in addition to ``s``.
        Default is True.
    overwrite_a : bool, optional
        Whether to overwrite data in `a` (may improve performance). Default is False.
        See :ref:`tutorial_linalg_overwrite` for details.
    check_finite : bool, optional
        Whether to check that the input matrix contains only finite numbers.
        Disabling may give a performance gain, but may result in problems
        (crashes, non-termination) if the inputs do contain infinities or NaNs.
    lapack_driver : {'gesdd', 'gesvd'}, optional
        Whether to use the more efficient divide-and-conquer approach
        (``'gesdd'``) or general rectangular approach (``'gesvd'``)
        to compute the SVD. MATLAB and Octave use the ``'gesvd'`` approach.
        Default is ``'gesdd'``.

    Returns
    -------
    U : ndarray
        Unitary matrix having left singular vectors as columns.
        Of shape ``(M, M)`` or ``(M, K)``, depending on `full_matrices`.
        Only present when ``compute_uv=True``.
    s : ndarray
        The singular values, sorted in non-increasing order.
        Of shape (K,), with ``K = min(M, N)``.
    Vh : ndarray
        Unitary matrix having right singular vectors as rows.
        Of shape ``(N, N)`` or ``(K, N)`` depending on `full_matrices`.
        Only present when ``compute_uv=True``.

    Raises
    ------
    LinAlgError
        If SVD computation does not converge.

    See Also
    --------
    svdvals : Compute singular values of a matrix.
    diagsvd : Construct the Sigma matrix, given the vector s.

    Notes
    -----
    The array argument of this function, `a`, may have additional
    "batch" dimensions prepended to the core shape. In this case, the array is treated
    as a batch of lower-dimensional slices; see :ref:`linalg_batch` for details.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy import linalg
    >>> rng = np.random.default_rng()
    >>> m, n = 9, 6
    >>> a = rng.standard_normal((m, n)) + 1.j*rng.standard_normal((m, n))
    >>> U, s, Vh = linalg.svd(a)
    >>> U.shape,  s.shape, Vh.shape
    ((9, 9), (6,), (6, 6))

    Reconstruct the original matrix from the decomposition:

    >>> sigma = np.zeros((m, n))
    >>> for i in range(min(m, n)):
    ...     sigma[i, i] = s[i]
    >>> a1 = U @ sigma @ Vh
    >>> np.allclose(a, a1)
    True

    Alternatively, use ``full_matrices=False`` (notice that the shape of
    ``U`` is then ``(m, n)`` instead of ``(m, m)``):

    >>> U, s, Vh = linalg.svd(a, full_matrices=False)
    >>> U.shape, s.shape, Vh.shape
    ((9, 6), (6,), (6, 6))
    >>> S = np.diag(s)
    >>> np.allclose(a, U @ S @ Vh)
    True

    >>> s2 = linalg.svd(a, compute_uv=False)
    >>> np.allclose(s, s2)
    True

    If the input matrix has more than two dimensions, it is interpreted as a batch of
    two-dimensional matrices:

    >>> aa = np.stack((a, 2*a))
    >>> linalg.svdvals(aa)[0] == linalg.svdvals(a)
    array([ True,  True,  True,  True,  True,  True])
    >>> linalg.svdvals(aa)[1] == 2 * linalg.svdvals(a)
    array([ True,  True,  True,  True,  True,  True])
    """
    if not isinstance(lapack_driver, str):
        raise TypeError('lapack_driver must be a string')
    if lapack_driver not in ('gesdd', 'gesvd'):
        message = f'lapack_driver must be "gesdd" or "gesvd", not "{lapack_driver}"'
        raise ValueError(message)

    # basic sanity checks of the input matrix
    a1 = _asarray_validated(a, check_finite=check_finite)
    _deprecate_dtypes("svd", a1)

    if a1.ndim < 2:
        raise ValueError(f"Expected at least ndim=2, got {a1.ndim=}")

    m, n = a1.shape[-2], a1.shape[-1]

    # Also check if dtype is LAPACK compatible
    a1, overwrite_a = _normalize_lapack_dtype(a1, overwrite_a)
    a1, overwrite_a = _ensure_aligned_and_native(a1, overwrite_a)

    overwrite_a = overwrite_a or (_datacopied(a1, a))
    overwrite_a = overwrite_a and (a1.ndim == 2) and (a1.flags["F_CONTIGUOUS"])

    # accommodate empty matrix
    if a1.size == 0:
        u0, s0, v0 = svd(np.eye(2, dtype=a1.dtype))

        batch_shape = a1.shape[:-2]
        s = np.empty_like(a1, shape=batch_shape + (0,), dtype=s0.dtype)
        if full_matrices:
            u = np.empty_like(a1, shape=batch_shape + (m, m), dtype=u0.dtype)
            u[...] = np.identity(m)
            v = np.empty_like(a1, shape=batch_shape + (n, n), dtype=v0.dtype)
            v[...] = np.identity(n)
        else:
            u = np.empty_like(a1, shape=batch_shape + (m, 0), dtype=u0.dtype)
            v = np.empty_like(a1, shape=batch_shape + (0, n), dtype=v0.dtype)
        if compute_uv:
            return u, s, v
        else:
            return s

    if compute_uv:
        max_mn, min_mn = (m, n) if m > n else (n, m)
        if full_matrices:
            if not HAS_ILP64 and max_mn*max_mn > np.iinfo(np.int32).max:
                raise ValueError(f"Indexing a matrix size {max_mn} x {max_mn} "
                                  "would incur integer overflow in LAPACK. "
                                  "Instead, either use using numpy.linalg.svd or build"
                                  "SciPy with ILP64 support.")
        else:
            sz = max(m * min_mn, n * min_mn)
            if not HAS_ILP64 and max(m * min_mn, n * min_mn) > np.iinfo(np.int32).max:
                raise ValueError(f"Indexing a matrix of {sz} elements would "
                                  "incur an in integer overflow in LAPACK. "
                                  "Instead, either use using numpy.linalg.svd or build"
                                  "SciPy with ILP64 support.")

    res = _batched_linalg._svd(
        a1, lapack_driver, compute_uv, full_matrices, overwrite_a
    )

    err_lst = res[-1]
    if err_lst:
        _format_emit_errors_warnings(err_lst, lapack_driver)

    if compute_uv:
        return res[:-1]    # u, s, v
    else:
        return res[0]   # s


def svdvals(a, overwrite_a=False, check_finite=True):
    """
    Compute singular values of a matrix.

    Parameters
    ----------
    a : (M, N) array_like
        Matrix to decompose.
    overwrite_a : bool, optional
        Whether to overwrite data in `a` (may improve performance). Default is False.
        See :ref:`tutorial_linalg_overwrite` for details.
    check_finite : bool, optional
        Whether to check that the input matrix contains only finite numbers.
        Disabling may give a performance gain, but may result in problems
        (crashes, non-termination) if the inputs do contain infinities or NaNs.

    Returns
    -------
    s : (min(M, N),) ndarray
        The singular values, sorted in decreasing order.

    Raises
    ------
    LinAlgError
        If SVD computation does not converge.

    See Also
    --------
    svd : Compute the full singular value decomposition of a matrix.
    diagsvd : Construct the Sigma matrix, given the vector s.

    Notes
    -----
    Array argument of this function, `a`, may have additional
    "batch" dimensions prepended to the core shape. In this case, the array is treated
    as a batch of lower-dimensional slices; see :ref:`linalg_batch` for details.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.linalg import svdvals
    >>> m = np.array([[1.0, 0.0],
    ...               [2.0, 3.0],
    ...               [1.0, 1.0],
    ...               [0.0, 2.0],
    ...               [1.0, 0.0]])
    >>> svdvals(m)
    array([ 4.28091555,  1.63516424])

    If the input matrix has more than two dimensions, it is interpreted as a batch of
    two-dimensional matrices:

    >>> mm = np.stack((m, 2*m))
    >>> svdvals(mm)
    array([[4.28091555, 1.63516424],
           [8.56183109, 3.27032847]])

    We can verify the maximum singular value of `m` by computing the maximum
    length of `m @ u` over all the unit vectors `u` in the (x,y) plane.
    We approximate "all" the unit vectors with a large sample. Because
    of linearity, we only need the unit vectors with angles in ``[0, pi]``.

    >>> t = np.linspace(0, np.pi, 2000)
    >>> u = np.array([np.cos(t), np.sin(t)])
    >>> np.linalg.norm(m @ u, axis=0).max()
    4.2809152422538475

    `p` is a projection matrix with rank 1. With exact arithmetic,
    its singular values would be ``[1, 0, 0, 0]``.

    >>> v = np.array([0.1, 0.3, 0.9, 0.3])
    >>> p = np.outer(v, v)
    >>> svdvals(p)
    array([  1.00000000e+00,   2.02021698e-17,   1.56692500e-17,
             8.15115104e-34])

    The singular values of an orthogonal matrix are all 1. Here, we
    create a random orthogonal matrix by using the ``rvs()`` method of
    `scipy.stats.ortho_group`.

    >>> from scipy.stats import ortho_group
    >>> orth = ortho_group.rvs(4)
    >>> svdvals(orth)
    array([ 1.,  1.,  1.,  1.])
    """
    return svd(a, compute_uv=0, overwrite_a=overwrite_a,
               check_finite=check_finite)


@_apply_over_batch(('s', 1))
def diagsvd(s, M, N):
    """
    Construct the sigma matrix in SVD from singular values and size M, N.

    Parameters
    ----------
    s : (M,) or (N,) array_like
        Singular values
    M : int
        Size of the matrix whose singular values are `s`.
    N : int
        Size of the matrix whose singular values are `s`.

    Returns
    -------
    S : (M, N) ndarray
        The S-matrix in the singular value decomposition

    See Also
    --------
    svd : Singular value decomposition of a matrix
    svdvals : Compute singular values of a matrix.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.linalg import diagsvd
    >>> vals = np.array([1, 2, 3])  # The array representing the computed svd
    >>> diagsvd(vals, 3, 4)
    array([[1, 0, 0, 0],
           [0, 2, 0, 0],
           [0, 0, 3, 0]])
    >>> diagsvd(vals, 4, 3)
    array([[1, 0, 0],
           [0, 2, 0],
           [0, 0, 3],
           [0, 0, 0]])

    """
    part = np.diag(s)
    typ = part.dtype.char
    MorN = len(s)
    if MorN == M:
        return np.hstack((part, np.zeros((M, N - M), dtype=typ)))
    elif MorN == N:
        return np.r_[part, np.zeros((M - N, N), dtype=typ)]
    else:
        raise ValueError("Length of s must be M or N.")


# Orthonormal decomposition

@_apply_over_batch(('A', 2))
def orth(A, rcond=None):
    """
    Construct an orthonormal basis for the range of A using SVD.

    Parameters
    ----------
    A : (M, N) array_like
        Input array
    rcond : float, optional
        Relative condition number. Singular values ``s`` smaller than
        ``rcond * max(s)`` are considered zero.
        Default: floating point eps * max(M,N).

    Returns
    -------
    Q : (M, K) ndarray
        Orthonormal basis for the range of A.
        K = effective rank of A, as determined by rcond

    See Also
    --------
    svd : Singular value decomposition of a matrix
    null_space : Matrix null space

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.linalg import orth
    >>> A = np.array([[2, 0, 0], [0, 5, 0]])  # rank 2 array
    >>> orth(A)
    array([[0., 1.],
           [1., 0.]])
    >>> orth(A.T)
    array([[0., 1.],
           [1., 0.],
           [0., 0.]])

    """
    u, s, vh = svd(A, full_matrices=False)
    M, N = u.shape[0], vh.shape[1]
    if rcond is None:
        rcond = np.finfo(s.dtype).eps * max(M, N)
    tol = np.amax(s, initial=0.) * rcond
    num = np.sum(s > tol, dtype=int)
    Q = u[:, :num]
    return Q


@_apply_over_batch(('A', 2))
def null_space(A, rcond=None, *, overwrite_a=False, check_finite=True,
               lapack_driver='gesdd'):
    """
    Construct an orthonormal basis for the null space of A using SVD.

    Parameters
    ----------
    A : (M, N) array_like
        Input array
    rcond : float, optional
        Relative condition number. Singular values ``s`` smaller than
        ``rcond * max(s)`` are considered zero.
        Default: floating point eps * max(M,N).
    overwrite_a : bool, optional
        Whether to overwrite `a`; may improve performance. Default is False.
        See :ref:`tutorial_linalg_overwrite` for details.
    check_finite : bool, optional
        Whether to check that the input matrix contains only finite numbers.
        Disabling may give a performance gain, but may result in problems
        (crashes, non-termination) if the inputs do contain infinities or NaNs.
    lapack_driver : {'gesdd', 'gesvd'}, optional
        Whether to use the more efficient divide-and-conquer approach
        (``'gesdd'``) or general rectangular approach (``'gesvd'``)
        to compute the SVD. MATLAB and Octave use the ``'gesvd'`` approach.
        Default is ``'gesdd'``.

    Returns
    -------
    Z : (N, K) ndarray
        Orthonormal basis for the null space of A.
        K = dimension of effective null space, as determined by rcond

    See Also
    --------
    svd : Singular value decomposition of a matrix
    orth : Matrix range

    Examples
    --------
    1-D null space:

    >>> import numpy as np
    >>> from scipy.linalg import null_space
    >>> A = np.array([[1, 1], [1, 1]])
    >>> ns = null_space(A)
    >>> ns * np.copysign(1, ns[0,0])  # Remove the sign ambiguity of the vector
    array([[ 0.70710678],
           [-0.70710678]])

    2-D null space:

    >>> from numpy.random import default_rng
    >>> rng = default_rng()
    >>> B = rng.random((3, 5))
    >>> Z = null_space(B)
    >>> Z.shape
    (5, 2)
    >>> np.allclose(B.dot(Z), 0)
    True

    The basis vectors are orthonormal (up to rounding error):

    >>> Z.T.dot(Z)
    array([[  1.00000000e+00,   6.92087741e-17],
           [  6.92087741e-17,   1.00000000e+00]])

    """
    u, s, vh = svd(A, full_matrices=True, overwrite_a=overwrite_a,
                   check_finite=check_finite, lapack_driver=lapack_driver)
    M, N = u.shape[0], vh.shape[1]
    if rcond is None:
        rcond = np.finfo(s.dtype).eps * max(M, N)
    tol = np.amax(s, initial=0.) * rcond
    num = np.sum(s > tol, dtype=int)
    Q = vh[num:,:].T.conj()
    return Q


@_apply_over_batch(('A', 2), ('B', 2))
def subspace_angles(A, B):
    r"""
    Compute the subspace angles between two matrices.

    Parameters
    ----------
    A : (M, N) array_like
        The first input array.
    B : (M, K) array_like
        The second input array.

    Returns
    -------
    angles : ndarray, shape (min(N, K),)
        The subspace angles between the column spaces of `A` and `B` in
        descending order.

    See Also
    --------
    orth
    svd

    Notes
    -----
    This computes the subspace angles according to the formula
    provided in [1]_. For equivalence with MATLAB and Octave behavior,
    use ``angles[0]``.

    .. versionadded:: 1.0

    References
    ----------
    .. [1] Knyazev A, Argentati M (2002) Principal Angles between Subspaces
           in an A-Based Scalar Product: Algorithms and Perturbation
           Estimates. SIAM J. Sci. Comput. 23:2008-2040.

    Examples
    --------
    A Hadamard matrix, which has orthogonal columns, so we expect that
    the subspace angle to be :math:`\frac{\pi}{2}`:

    >>> import numpy as np
    >>> from scipy.linalg import hadamard, subspace_angles
    >>> rng = np.random.default_rng()
    >>> H = hadamard(4)
    >>> print(H)
    [[ 1  1  1  1]
     [ 1 -1  1 -1]
     [ 1  1 -1 -1]
     [ 1 -1 -1  1]]
    >>> np.rad2deg(subspace_angles(H[:, :2], H[:, 2:]))
    array([ 90.,  90.])

    And the subspace angle of a matrix to itself should be zero:

    >>> subspace_angles(H[:, :2], H[:, :2]) <= 2 * np.finfo(float).eps
    array([ True,  True], dtype=bool)

    The angles between non-orthogonal subspaces are in between these extremes:

    >>> x = rng.standard_normal((4, 3))
    >>> np.rad2deg(subspace_angles(x[:, :2], x[:, [2]]))
    array([ 55.832])  # random
    """
    # Steps here omit the U and V calculation steps from the paper

    # 1. Compute orthonormal bases of column-spaces
    A = _asarray_validated(A, check_finite=True)
    if len(A.shape) != 2:
        raise ValueError(f'expected 2D array, got shape {A.shape}')
    QA = orth(A)
    del A

    B = _asarray_validated(B, check_finite=True)
    if len(B.shape) != 2:
        raise ValueError(f'expected 2D array, got shape {B.shape}')
    if len(B) != len(QA):
        raise ValueError('A and B must have the same number of rows, got '
                         f'{QA.shape[0]} and {B.shape[0]}')
    QB = orth(B)
    del B

    # 2. Compute SVD for cosine
    QA_H_QB = np.dot(QA.T.conj(), QB)
    sigma = svdvals(QA_H_QB)

    # 3. Compute matrix B
    if QA.shape[1] >= QB.shape[1]:
        B = QB - np.dot(QA, QA_H_QB)
    else:
        B = QA - np.dot(QB, QA_H_QB.T.conj())
    del QA, QB, QA_H_QB

    # 4. Compute SVD for sine
    mask = sigma ** 2 >= 0.5
    if mask.any():
        mu_arcsin = np.arcsin(np.clip(svdvals(B, overwrite_a=True), -1., 1.))
    else:
        mu_arcsin = 0.

    # 5. Compute the principal angles
    # with reverse ordering of sigma because smallest sigma belongs to largest
    # angle theta
    theta = np.where(mask, mu_arcsin, np.arccos(np.clip(sigma[::-1], -1., 1.)))
    return theta



"""
This module is home to specific dtypes related functionality and their classes.
For more general information about dtypes, also see `numpy.dtype` and
:ref:`arrays.dtypes`.

Similar to the builtin ``types`` module, this submodule defines types (classes)
that are not widely used directly.

.. versionadded:: NumPy 1.25

    The dtypes module is new in NumPy 1.25.  Previously DType classes were
    only accessible indirectly.


DType classes
-------------

The following are the classes of the corresponding NumPy dtype instances and
NumPy scalar types.  The classes can be used in ``isinstance`` checks and can
also be instantiated or used directly.  Direct use of these classes is not
typical, since their scalar counterparts (e.g. ``np.float64``) or strings
like ``"float64"`` can be used.
"""

# See doc/source/reference/routines.dtypes.rst for module-level docs
__all__ = ["register_dlpack_dtype"]


def register_dlpack_dtype(dlpack_key, dtype, /):
    """
    Register a NumPy dtype for a DLPack ``(code, bits)`` pair so that
    `numpy.from_dlpack` can import it and ``ndarray.__dlpack__`` can export
    it.  Built-in dtype mappings take priority on import.

    If you think a conflict is possible but unproblematic you may wrap this
    into a try/except block as NumPy will raise an error if another DType
    is already registered for the same (code, bits) pair.
    While an error is raised, the export is registered even on error.

    Registering an identical (not equal) dtype multiple times is allowed but
    normally registration should happen at import time.

    .. warning::
        It is the responsibility of the registering user to ensure that the
        mapping is valid.

    .. note::
        This function was added primarily for ``ml_dtypes`` and may be
        replaced with a different mechanism in the future.  It is intended
        to be used by authors of user-defined dtypes and not end-users.

    Parameters
    ----------
    dlpack_key : tuple of int
        ``(dl_dtype_code, dl_bits)`` matching the DLPack ``DLDataType``,
        lanes is assumed to be always 1, currently.
    dtype : dtype
        A NumPy dtype instance.

    Raises
    ------
    ValueError : If a conflicting registration was already done.

    See Also
    --------
    numpy.from_dlpack

    """
    # Deferred to avoid circular import.
    from numpy._core._multiarray_umath import _register_dlpack_dtype

    return _register_dlpack_dtype(dlpack_key, dtype)


def _add_dtype_helper(DType, alias):
    # Function to add DTypes a bit more conveniently without channeling them
    # through `numpy._core._multiarray_umath` namespace or similar.
    from numpy import dtypes

    setattr(dtypes, DType.__name__, DType)
    __all__.append(DType.__name__)

    if alias:
        alias = alias.removeprefix("numpy.dtypes.")
        setattr(dtypes, alias, DType)
        __all__.append(alias)



"""Input validation helpers."""

from __future__ import annotations

import math


def validate_count(name: str, value: int) -> None:
    """Validate a non-negative integer count."""
    if isinstance(value, bool) or not isinstance(value, int):
        raise ValueError(f"{name} must be an integer, got {value!r}.")
    if value < 0:
        raise ValueError(f"{name} must be >= 0, got {value}.")


def validate_total(name: str, value: int) -> None:
    """Validate a positive integer total."""
    if isinstance(value, bool) or not isinstance(value, int):
        raise ValueError(f"{name} must be an integer, got {value!r}.")
    if value <= 0:
        raise ValueError(f"{name} must be > 0, got {value}.")


def validate_successes_le_total(
    successes_name: str,
    successes: int,
    total_name: str,
    total: int,
) -> None:
    """Validate that successes do not exceed total."""
    if successes > total:
        raise ValueError(
            f"{successes_name} ({successes}) cannot exceed {total_name} ({total})."
        )


def validate_prior(name: str, value: float) -> None:
    """Validate a positive Beta prior parameter."""
    if isinstance(value, bool) or not isinstance(value, (int, float)):
        raise ValueError(f"{name} must be a number, got {value!r}.")
    if value <= 0:
        raise ValueError(f"{name} must be > 0, got {value}.")


def validate_credible_interval(value: float) -> None:
    """Validate a credible interval level between 0 and 1."""
    if isinstance(value, bool) or not isinstance(value, (int, float)):
        raise ValueError(f"credible_interval must be a number, got {value!r}.")
    if not (0.0 < value < 1.0):
        raise ValueError(
            f"credible_interval must be strictly between 0 and 1, got {value}."
        )


def validate_n_simulations(value: int) -> None:
    """Validate a positive integer number of simulations."""
    if isinstance(value, bool) or not isinstance(value, int):
        raise ValueError(f"n_simulations must be an integer, got {value!r}.")
    if value <= 0:
        raise ValueError(f"n_simulations must be > 0, got {value}.")


def validate_seed(value: int | None) -> None:
    """Validate a random seed."""
    if value is None:
        return
    if isinstance(value, bool) or not isinstance(value, int):
        raise ValueError(f"seed must be None or an integer, got {value!r}.")
    if value < 0:
        raise ValueError(f"seed must be >= 0, got {value}.")


def validate_margin(value: float) -> None:
    """Validate a non-negative, finite do-no-harm margin."""
    if isinstance(value, bool) or not isinstance(value, (int, float)):
        raise ValueError(f"margin must be a number, got {value!r}.")
    if not math.isfinite(value):
        raise ValueError(f"margin must be finite, got {value}.")
    if value < 0:
        raise ValueError(f"margin must be >= 0, got {value}.")


"""
=============================================================
Spatial algorithms and data structures (:mod:`scipy.spatial`)
=============================================================

.. currentmodule:: scipy.spatial

.. toctree::
   :hidden:

   spatial.distance
   spatial.transform

Spatial transformations
=======================

These are contained in the `scipy.spatial.transform` submodule.

Nearest-neighbor queries
========================
.. autosummary::
   :toctree: generated/

   KDTree      -- class for efficient nearest-neighbor queries
   cKDTree     -- class for efficient nearest-neighbor queries (faster implementation)
   Rectangle

Distance metrics
================

Distance metrics are contained in the :mod:`scipy.spatial.distance` submodule.

Delaunay triangulation, convex hulls, and Voronoi diagrams
==========================================================

.. autosummary::
   :toctree: generated/

   Delaunay    -- compute Delaunay triangulation of input points
   ConvexHull  -- compute a convex hull for input points
   Voronoi     -- compute a Voronoi diagram hull from input points
   SphericalVoronoi -- compute a Voronoi diagram from input points on the surface of a sphere
   HalfspaceIntersection -- compute the intersection points of input halfspaces

Plotting helpers
================

.. autosummary::
   :toctree: generated/

   delaunay_plot_2d     -- plot 2-D triangulation
   convex_hull_plot_2d  -- plot 2-D convex hull
   voronoi_plot_2d      -- plot 2-D Voronoi diagram

.. seealso:: :ref:`Tutorial <qhulltutorial>`


Simplex representation
======================
The simplices (triangles, tetrahedra, etc.) appearing in the Delaunay
tessellation (N-D simplices), convex hull facets, and Voronoi ridges
(N-1-D simplices) are represented in the following scheme::

    tess = Delaunay(points)
    hull = ConvexHull(points)
    voro = Voronoi(points)

    # coordinates of the jth vertex of the ith simplex
    tess.points[tess.simplices[i, j], :]        # tessellation element
    hull.points[hull.simplices[i, j], :]        # convex hull facet
    voro.vertices[voro.ridge_vertices[i, j], :] # ridge between Voronoi cells

For Delaunay triangulations and convex hulls, the neighborhood
structure of the simplices satisfies the condition:
``tess.neighbors[i,j]`` is the neighboring simplex of the ith
simplex, opposite to the ``j``-vertex. It is -1 in case of no neighbor.

Convex hull facets also define a hyperplane equation::

    (hull.equations[i,:-1] * coord).sum() + hull.equations[i,-1] == 0

Similar hyperplane equations for the Delaunay triangulation correspond
to the convex hull facets on the corresponding N+1-D
paraboloid.

The Delaunay triangulation objects offer a method for locating the
simplex containing a given point, and barycentric coordinate
computations.

Miscellaneous Functions
=======================

.. autosummary::
   :toctree: generated/

   tsearch
   distance_matrix
   minkowski_distance
   minkowski_distance_p
   procrustes
   geometric_slerp

Warnings / Errors used in :mod:`scipy.spatial`
==============================================
.. autosummary::
   :toctree: generated/

   QhullError
"""  # noqa: E501

from ._kdtree import *
from ._ckdtree import *  # type: ignore[import-not-found]
from ._qhull import *
from ._spherical_voronoi import SphericalVoronoi
from ._plotutils import *
from ._procrustes import procrustes
from ._geometric_slerp import geometric_slerp

# Deprecated namespaces, to be removed in v2.0.0
from . import ckdtree, kdtree, qhull

__all__ = [s for s in dir() if not s.startswith('_')]

from . import distance, transform

__all__ += ['distance', 'transform']

from scipy._lib._testutils import PytestTester
test = PytestTester(__name__)
del PytestTester



"""Sparse Bit Set encoding/decoding for IFT (Incremental Font Transfer).

Implements the sparse bit set format defined in the W3C IFT specification:
https://w3c.github.io/IFT/Overview.html#sparse-bit-set-decoding
"""

# Reference: https://github.com/googlefonts/fontations/blob/main/read-fonts/src/collections/int_set/sparse_bit_set.rs


import bisect
from collections import deque
from typing import Dict, Iterable, Optional, Set, Tuple

# Maximum tree height that fits within a 32-bit leaf node for each branching factor.
_BF_MAX_HEIGHT: Dict[int, int] = {2: 31, 4: 16, 8: 11, 32: 7}


class SparseBitSetDecodeError(Exception):
    pass


def decode(
    data: bytes, bias: int = 0, maxValue: int = 0xFFFFFFFF
) -> Tuple[Set[int], int]:
    """Decode a sparse bit set from binary data.

    Args:
        data: bytes-like object containing the sparse bit set encoding.
        bias: integer added to each decoded value.

    Returns:
        A tuple (values, bytesConsumed) where values is a set of integers
        and bytesConsumed is the number of bytes read from data.
    """
    if not data:
        raise SparseBitSetDecodeError("Empty data")

    branchFactor, height = _decodeHeader(data[0])

    maxHeight = _BF_MAX_HEIGHT[branchFactor]
    if height > maxHeight:
        raise SparseBitSetDecodeError(
            f"Height {height} exceeds max {maxHeight} for branch factor {branchFactor}"
        )

    return _decodeImpl(data, branchFactor, height, bias, maxValue)


def _decodeImpl(
    data: bytes, branchFactor: int, height: int, bias: int, maxValue: int
) -> Tuple[Set[int], int]:
    if height == 0:
        # 1 byte was used for the header.
        return (set(), 1)

    bitStream = _InputBitStream(data, branchFactor)
    result: Set[int] = set()
    # Queue entries are (startValue, depth), where startValue is the first
    # integer that could be covered by this node.
    queue: deque[Tuple[int, int]] = deque()
    queue.append((0, 1))

    while queue:
        start, depth = queue.popleft()
        bits = bitStream.next()
        if bits is None:
            raise SparseBitSetDecodeError("Unexpected end of data")

        # all bits were were zero which is a special command to completely fill
        # in all integers covered by this node.
        if bits == 0:
            exp = height - depth + 1
            nodeSize = branchFactor**exp
            fillStart = start + bias
            if fillStart > maxValue:
                continue
            fillEnd = min(maxValue, start + nodeSize - 1 + bias)
            if fillStart < 0:
                fillStart = 0
            if fillStart <= fillEnd:
                result.update(range(fillStart, fillEnd + 1))
            continue

        # Non-zero node: each set bit identifies a child/value.
        exp = height - depth
        nextNodeSize = branchFactor**exp
        while True:
            bitIndex = _trailingZeros(bits, 32)
            if bitIndex == 32:
                break
            if depth == height:
                val = start + bitIndex + bias
                if val > maxValue:
                    queue.clear()
                    break
                if val >= 0:
                    result.add(val)
            else:
                startDelta = bitIndex * nextNodeSize
                queue.append((start + startDelta, depth + 1))
            bits &= ~(1 << bitIndex)

    return (result, bitStream.bytesConsumed())


def encode(values: Iterable[int]) -> bytes:
    """Encode a set of integers as a sparse bit set.

    Tries all branching factors and returns the shortest encoding.

    Args:
        values: iterable of non-negative integers.

    Returns:
        bytes containing the sparse bit set encoding.
    """
    valuesSorted = sorted(set(values))
    if not valuesSorted:
        return _encodeHeader(2, 0)

    maxValue = valuesSorted[-1]
    valueSet = set(valuesSorted)

    best: Optional[bytes] = None
    for branchFactor in (2, 4, 8, 32):
        height = _treeHeight(branchFactor, maxValue)
        if height > _BF_MAX_HEIGHT[branchFactor]:
            continue
        encoded = _encodeWithBf(valueSet, branchFactor, height)
        if best is None or len(encoded) < len(best):
            best = encoded

    if best is None:
        raise ValueError(f"Cannot encode max value {maxValue}")
    return best


def _encodeHeader(branchFactor: int, height: int) -> bytes:
    branchFactorToId = {2: 0, 4: 1, 8: 2, 32: 3}
    return bytes([(height << 2) | branchFactorToId[branchFactor]])


def _decodeHeader(headerByte: int) -> Tuple[int, int]:
    id = headerByte & 0x03
    idToBranchFactor = {0: 2, 1: 4, 2: 8, 3: 32}
    height = (headerByte >> 2) & 0x1F
    return idToBranchFactor[id], height


def _treeHeight(branchFactor: int, maxValue: int) -> int:
    """Return the minimum tree height needed to represent maxValue."""
    height = 1
    capacity = branchFactor
    while capacity <= maxValue:
        capacity *= branchFactor
        height += 1
    return height


def _encodeWithBf(valueSet: Set[int], branchFactor: int, height: int) -> bytes:
    if height == 0:
        return _encodeHeader(branchFactor, 0)

    # Build layers bottom-up: layer 0 = leaves (individual values),
    # each higher layer groups bf children into one parent bitmask.
    layers: list[Dict[int, int]] = [{}]  # list of dicts: nodeIndex -> bitmask

    for v in valueSet:
        nodeIndex = v // branchFactor
        bitPos = v % branchFactor
        if nodeIndex not in layers[0]:
            layers[0][nodeIndex] = 0
        layers[0][nodeIndex] |= 1 << bitPos

    for _ in range(1, height):
        prevLayer = layers[-1]
        newLayer: Dict[int, int] = {}
        for nodeIndex, bitmask in prevLayer.items():
            parentIndex = nodeIndex // branchFactor
            bitPos = nodeIndex % branchFactor
            if parentIndex not in newLayer:
                newLayer[parentIndex] = 0
            newLayer[parentIndex] |= 1 << bitPos
        layers.append(newLayer)

    # For zero-node optimization: track count of values in sorted list.
    valuesSorted = sorted(valueSet)

    def rangeCount(lo: int, hi: int) -> int:
        return bisect.bisect_right(valuesSorted, hi) - bisect.bisect_left(
            valuesSorted, lo
        )

    # Emit nodes BFS order (root to leaves).
    # Queue entries: (nodeIndex, depthFromRoot, rangeStart, rangeEnd)
    stream = _OutputBitStream(branchFactor)
    subtreeSize = branchFactor**height
    queue: deque[Tuple[int, int, int, int]] = deque([(0, 0, 0, subtreeSize - 1)])

    while queue:
        nodeIndex, depth, rangeStart, rangeEnd = queue.popleft()
        layerIdx = height - 1 - depth  # layers[0]=leaves, layers[height-1]=root

        bitmask = (
            layers[layerIdx].get(nodeIndex, 0) if 0 <= layerIdx < len(layers) else 0
        )

        # Zero-node optimization: if entire range is filled on an INTERNAL node,
        # write 0 and skip children.  At leaf level we always write the explicit
        # bitmask so the encoding matches the reference.
        if (
            depth < height - 1
            and rangeCount(rangeStart, rangeEnd) == rangeEnd - rangeStart + 1
        ):
            stream.write(0)
            continue

        stream.write(bitmask)

        if bitmask != 0 and depth < height - 1:
            childSize = (rangeEnd - rangeStart + 1) // branchFactor
            bits = bitmask
            while bits:
                bitIndex = _trailingZeros(bits, 32)
                childIndex = nodeIndex * branchFactor + bitIndex
                childStart = rangeStart + bitIndex * childSize
                childEnd = childStart + childSize - 1
                queue.append((childIndex, depth + 1, childStart, childEnd))
                bits &= ~(1 << bitIndex)

    return _encodeHeader(branchFactor, height) + stream.toBytes()


def _trailingZeros(val: int, maxBits: int) -> int:
    if val == 0:
        return maxBits
    count = 0
    while (val & 1) == 0:
        val >>= 1
        count += 1
    return count


class _InputBitStream:
    """Reads bit nodes from a byte array, starting after the header byte."""

    def __init__(self, data: bytes, branchFactor: int):
        self.data = data
        self.branchFactor = branchFactor
        self.byteIndex = 1  # skip the header byte
        self.subIndex = 0  # bit offset within current byte (for bf=2,4)

    def next(self) -> Optional[int]:
        if self.branchFactor in (2, 4):
            if self.byteIndex >= len(self.data):
                return None
            mask = (1 << self.branchFactor) - 1
            val = (self.data[self.byteIndex] >> self.subIndex) & mask
            self.subIndex += self.branchFactor
            if self.subIndex >= 8:
                self.subIndex = 0
                self.byteIndex += 1
            return val
        elif self.branchFactor == 8:
            if self.byteIndex >= len(self.data):
                return None
            val = self.data[self.byteIndex]
            self.byteIndex += 1
            return val
        elif self.branchFactor == 32:
            if self.byteIndex + 3 >= len(self.data):
                return None
            b = self.data
            i = self.byteIndex
            val = b[i] | (b[i + 1] << 8) | (b[i + 2] << 16) | (b[i + 3] << 24)
            self.byteIndex += 4
            return val
        return None

    def bytesConsumed(self) -> int:
        return self.byteIndex + (1 if self.subIndex > 0 else 0)


class _OutputBitStream:
    """Writes bit nodes into a byte array."""

    def __init__(self, branchFactor: int):
        self.branchFactor = branchFactor
        self.data = bytearray()
        self.subIndex = 0  # bit offset within current byte (for bf=2,4)

    def write(self, value: int) -> None:
        if self.branchFactor in (2, 4):
            mask = (1 << self.branchFactor) - 1
            value &= mask
            if self.subIndex == 0:
                self.data.append(0)
            self.data[-1] |= value << self.subIndex
            self.subIndex += self.branchFactor
            if self.subIndex >= 8:
                self.subIndex = 0
        elif self.branchFactor == 8:
            self.data.append(value & 0xFF)
        elif self.branchFactor == 32:
            self.data.append(value & 0xFF)
            self.data.append((value >> 8) & 0xFF)
            self.data.append((value >> 16) & 0xFF)
            self.data.append((value >> 24) & 0xFF)

    def toBytes(self) -> bytes:
        return bytes(self.data)



from .otBase import BaseTTXConverter


class table_H_V_A_R_(BaseTTXConverter):
    """Horizontal Metrics Variations table

    The ``HVAR`` table contains variations in horizontal glyph metrics
    in variable fonts.

    See also https://learn.microsoft.com/en-us/typography/opentype/spec/hvar
    """

    pass



from numpy.typing import NDArray
from typing import Any

def issymmetric(
    a: NDArray[Any],
    atol: None | float = ...,
    rtol: None | float = ...,
) -> bool: ...

def ishermitian(
    a: NDArray[Any],
    atol: None | float = ...,
    rtol: None | float = ...,
) -> bool: ...



import ctypes as ct
from _typeshed import StrOrBytesPath
from collections.abc import Iterable, Sequence
from typing import Any, ClassVar, Literal as L, overload

import numpy as np
from numpy._core._internal import _ctypes
from numpy._core.multiarray import flagsobj
from numpy._typing import (
    DTypeLike,
    NDArray,
    _AnyShape,
    _ArrayLike,
    _BoolCodes,
    _DTypeLike,
    _Float32Codes,
    _Float64Codes,
    _Int8Codes,
    _Int16Codes,
    _Int32Codes,
    _Int64Codes,
    _IntCCodes,
    _IntPCodes,
    _LongCodes,
    _LongDoubleCodes,
    _LongLongCodes,
    _ShapeLike,
    _UInt8Codes,
    _UInt16Codes,
    _UInt32Codes,
    _UInt64Codes,
    _UIntCCodes,
    _UIntPCodes,
    _ULongCodes,
    _ULongLongCodes,
    _VoidDTypeLike,
)

__all__ = ["load_library", "ndpointer", "c_intp", "as_ctypes", "as_array", "as_ctypes_type"]

type _FlagsKind = L[
    "C_CONTIGUOUS", "CONTIGUOUS", "C",
    "F_CONTIGUOUS", "FORTRAN", "F",
    "ALIGNED", "A",
    "WRITEABLE", "W",
    "OWNDATA", "O",
    "WRITEBACKIFCOPY", "X",
]

# TODO: Add a shape type parameter
class _ndptr[OptionalDTypeT: np.dtype | None](ct.c_void_p):
    # In practice these 4 classvars are defined in the dynamic class
    # returned by `ndpointer`
    _dtype_: OptionalDTypeT = ...
    _shape_: ClassVar[_AnyShape | None] = ...
    _ndim_: ClassVar[int | None] = ...
    _flags_: ClassVar[list[_FlagsKind] | None] = ...

    @overload  # type: ignore[override]
    @classmethod
    def from_param(cls: type[_ndptr[None]], obj: np.ndarray) -> _ctypes[Any]: ...  # pyrefly: ignore[bad-override]
    @overload
    @classmethod
    def from_param[DTypeT: np.dtype](cls: type[_ndptr[DTypeT]], obj: np.ndarray[Any, DTypeT]) -> _ctypes[Any]: ...  # pyright: ignore[reportIncompatibleMethodOverride]

class _concrete_ndptr[DTypeT: np.dtype](_ndptr[DTypeT]):
    _dtype_: DTypeT = ...
    _shape_: ClassVar[_AnyShape] = ...  # pyright: ignore[reportIncompatibleVariableOverride]

    @property
    def contents(self) -> np.ndarray[_AnyShape, DTypeT]: ...

def load_library(libname: StrOrBytesPath, loader_path: StrOrBytesPath) -> ct.CDLL: ...

c_intp = ct.c_int64  # most platforms are 64-bit nowadays

#
@overload
def ndpointer(
    dtype: None = None,
    ndim: int | None = None,
    shape: _ShapeLike | None = None,
    flags: _FlagsKind | Iterable[_FlagsKind] | int | flagsobj | None = None,
) -> type[_ndptr[None]]: ...
@overload
def ndpointer[ScalarT: np.generic](
    dtype: _DTypeLike[ScalarT],
    ndim: int | None = None,
    *,
    shape: _ShapeLike,
    flags: _FlagsKind | Iterable[_FlagsKind] | int | flagsobj | None = None,
) -> type[_concrete_ndptr[np.dtype[ScalarT]]]: ...
@overload
def ndpointer(
    dtype: DTypeLike | None,
    ndim: int | None = None,
    *,
    shape: _ShapeLike,
    flags: _FlagsKind | Iterable[_FlagsKind] | int | flagsobj | None = None,
) -> type[_concrete_ndptr[np.dtype]]: ...
@overload
def ndpointer[ScalarT: np.generic](
    dtype: _DTypeLike[ScalarT],
    ndim: int | None = None,
    shape: None = None,
    flags: _FlagsKind | Iterable[_FlagsKind] | int | flagsobj | None = None,
) -> type[_ndptr[np.dtype[ScalarT]]]: ...
@overload
def ndpointer(
    dtype: DTypeLike | None,
    ndim: int | None = None,
    shape: None = None,
    flags: _FlagsKind | Iterable[_FlagsKind] | int | flagsobj | None = None,
) -> type[_ndptr[np.dtype]]: ...

#
@overload  # bool
def as_ctypes_type(dtype: _BoolCodes | _DTypeLike[np.bool] | type[ct.c_bool]) -> type[ct.c_bool]: ...
@overload  # int8
def as_ctypes_type(dtype: _Int8Codes | _DTypeLike[np.int8] | type[ct.c_int8]) -> type[ct.c_int8]: ...
@overload  # int16
def as_ctypes_type(dtype: _Int16Codes | _DTypeLike[np.int16] | type[ct.c_int16]) -> type[ct.c_int16]: ...
@overload  # int32
def as_ctypes_type(dtype: _Int32Codes | _DTypeLike[np.int32] | type[ct.c_int32]) -> type[ct.c_int32]: ...
@overload  # int64
def as_ctypes_type(dtype: _Int64Codes | _DTypeLike[np.int64] | type[ct.c_int64]) -> type[ct.c_int64]: ...
@overload  # intc
def as_ctypes_type(dtype: _IntCCodes | type[ct.c_int]) -> type[ct.c_int]: ...
@overload  # long
def as_ctypes_type(dtype: _LongCodes | type[ct.c_long]) -> type[ct.c_long]: ...
@overload  # longlong
def as_ctypes_type(dtype: _LongLongCodes | type[ct.c_longlong]) -> type[ct.c_longlong]: ...
@overload  # intp
def as_ctypes_type(dtype: _IntPCodes | type[ct.c_ssize_t] | type[int]) -> type[ct.c_ssize_t]: ...
@overload  # uint8
def as_ctypes_type(dtype: _UInt8Codes | _DTypeLike[np.uint8] | type[ct.c_uint8]) -> type[ct.c_uint8]: ...
@overload  # uint16
def as_ctypes_type(dtype: _UInt16Codes | _DTypeLike[np.uint16] | type[ct.c_uint16]) -> type[ct.c_uint16]: ...
@overload  # uint32
def as_ctypes_type(dtype: _UInt32Codes | _DTypeLike[np.uint32] | type[ct.c_uint32]) -> type[ct.c_uint32]: ...
@overload  # uint64
def as_ctypes_type(dtype: _UInt64Codes | _DTypeLike[np.uint64] | type[ct.c_uint64]) -> type[ct.c_uint64]: ...
@overload  # uintc
def as_ctypes_type(dtype: _UIntCCodes | type[ct.c_uint]) -> type[ct.c_uint]: ...
@overload  # ulong
def as_ctypes_type(dtype: _ULongCodes | type[ct.c_ulong]) -> type[ct.c_ulong]: ...
@overload  # ulonglong
def as_ctypes_type(dtype: _ULongLongCodes | type[ct.c_ulonglong]) -> type[ct.c_ulonglong]: ...
@overload  # uintp
def as_ctypes_type(dtype: _UIntPCodes | type[ct.c_size_t]) -> type[ct.c_size_t]: ...
@overload  # float32
def as_ctypes_type(dtype: _Float32Codes | _DTypeLike[np.float32] | type[ct.c_float]) -> type[ct.c_float]: ...
@overload  # float64
def as_ctypes_type(dtype: _Float64Codes | _DTypeLike[np.float64] | type[float | ct.c_double]) -> type[ct.c_double]: ...
@overload  # longdouble
def as_ctypes_type(dtype: _LongDoubleCodes | _DTypeLike[np.longdouble] | type[ct.c_longdouble]) -> type[ct.c_longdouble]: ...
@overload  # void
def as_ctypes_type(dtype: _VoidDTypeLike) -> type[Any]: ...  # `ct.Union` or `ct.Structure`
@overload  # fallback
def as_ctypes_type(dtype: str) -> type[Any]: ...

#
@overload
def as_array(obj: ct._PointerLike, shape: Sequence[int]) -> NDArray[Any]: ...
@overload
def as_array[ScalarT: np.generic](obj: _ArrayLike[ScalarT], shape: _ShapeLike | None = None) -> NDArray[ScalarT]: ...
@overload
def as_array(obj: object, shape: _ShapeLike | None = None) -> NDArray[Any]: ...

#
@overload
def as_ctypes(obj: np.bool) -> ct.c_bool: ...
@overload
def as_ctypes(obj: np.int8) -> ct.c_int8: ...
@overload
def as_ctypes(obj: np.int16) -> ct.c_int16: ...
@overload
def as_ctypes(obj: np.int32) -> ct.c_int32: ...
@overload
def as_ctypes(obj: np.int64) -> ct.c_int64: ...
@overload
def as_ctypes(obj: np.uint8) -> ct.c_uint8: ...
@overload
def as_ctypes(obj: np.uint16) -> ct.c_uint16: ...
@overload
def as_ctypes(obj: np.uint32) -> ct.c_uint32: ...
@overload
def as_ctypes(obj: np.uint64) -> ct.c_uint64: ...
@overload
def as_ctypes(obj: np.float32) -> ct.c_float: ...
@overload
def as_ctypes(obj: np.float64) -> ct.c_double: ...
@overload
def as_ctypes(obj: np.longdouble) -> ct.c_longdouble: ...
@overload
def as_ctypes(obj: np.void) -> Any: ...  # `ct.Union` or `ct.Structure`
@overload
def as_ctypes(obj: NDArray[np.bool]) -> ct.Array[ct.c_bool]: ...
@overload
def as_ctypes(obj: NDArray[np.int8]) -> ct.Array[ct.c_int8]: ...
@overload
def as_ctypes(obj: NDArray[np.int16]) -> ct.Array[ct.c_int16]: ...
@overload
def as_ctypes(obj: NDArray[np.int32]) -> ct.Array[ct.c_int32]: ...
@overload
def as_ctypes(obj: NDArray[np.int64]) -> ct.Array[ct.c_int64]: ...
@overload
def as_ctypes(obj: NDArray[np.uint8]) -> ct.Array[ct.c_uint8]: ...
@overload
def as_ctypes(obj: NDArray[np.uint16]) -> ct.Array[ct.c_uint16]: ...
@overload
def as_ctypes(obj: NDArray[np.uint32]) -> ct.Array[ct.c_uint32]: ...
@overload
def as_ctypes(obj: NDArray[np.uint64]) -> ct.Array[ct.c_uint64]: ...
@overload
def as_ctypes(obj: NDArray[np.float32]) -> ct.Array[ct.c_float]: ...
@overload
def as_ctypes(obj: NDArray[np.float64]) -> ct.Array[ct.c_double]: ...
@overload
def as_ctypes(obj: NDArray[np.longdouble]) -> ct.Array[ct.c_longdouble]: ...
@overload
def as_ctypes(obj: NDArray[np.void]) -> ct.Array[Any]: ...  # `ct.Union` or `ct.Structure`



# Since bitmap glyph metrics are shared between EBLC and EBDT
# this class gets its own python file.
from fontTools.misc import sstruct
from fontTools.misc.textTools import safeEval
import logging


log = logging.getLogger(__name__)

bigGlyphMetricsFormat = """
  > # big endian
  height:       B
  width:        B
  horiBearingX: b
  horiBearingY: b
  horiAdvance:  B
  vertBearingX: b
  vertBearingY: b
  vertAdvance:  B
"""

smallGlyphMetricsFormat = """
  > # big endian
  height:   B
  width:    B
  BearingX: b
  BearingY: b
  Advance:  B
"""


class BitmapGlyphMetrics(object):
    def toXML(self, writer, ttFont):
        writer.begintag(self.__class__.__name__)
        writer.newline()
        for metricName in sstruct.getformat(self.__class__.binaryFormat)[1]:
            writer.simpletag(metricName, value=getattr(self, metricName))
            writer.newline()
        writer.endtag(self.__class__.__name__)
        writer.newline()

    def fromXML(self, name, attrs, content, ttFont):
        metricNames = set(sstruct.getformat(self.__class__.binaryFormat)[1])
        for element in content:
            if not isinstance(element, tuple):
                continue
            name, attrs, content = element
            # Make sure this is a metric that is needed by GlyphMetrics.
            if name in metricNames:
                vars(self)[name] = safeEval(attrs["value"])
            else:
                log.warning(
                    "unknown name '%s' being ignored in %s.",
                    name,
                    self.__class__.__name__,
                )


class BigGlyphMetrics(BitmapGlyphMetrics):
    binaryFormat = bigGlyphMetricsFormat


class SmallGlyphMetrics(BitmapGlyphMetrics):
    binaryFormat = smallGlyphMetricsFormat



from collections.abc import Sequence
from typing import Any, NotRequired, Protocol, TypedDict, runtime_checkable

import numpy as np

from ._char_codes import (
    _BoolCodes,
    _BytesCodes,
    _ComplexFloatingCodes,
    _DT64Codes,
    _FloatingCodes,
    _NumberCodes,
    _ObjectCodes,
    _SignedIntegerCodes,
    _StrCodes,
    _TD64Codes,
    _UnsignedIntegerCodes,
    _VoidCodes,
)

type _DTypeLikeNested = Any  # TODO: wait for support for recursive types


class _DTypeDict(TypedDict):
    names: Sequence[str]
    formats: Sequence[_DTypeLikeNested]
    # Only `str` elements are usable as indexing aliases,
    # but `titles` can in principle accept any object
    offsets: NotRequired[Sequence[int]]
    titles: NotRequired[Sequence[Any]]
    itemsize: NotRequired[int]
    aligned: NotRequired[bool]


# A protocol for anything with the dtype attribute
@runtime_checkable
class _HasDType[DTypeT: np.dtype](Protocol):
    @property
    def dtype(self) -> DTypeT: ...


@runtime_checkable
class _HasNumPyDType[DTypeT: np.dtype](Protocol):
    @property
    def __numpy_dtype__(self, /) -> DTypeT: ...


type _SupportsDType[DTypeT: np.dtype] = _HasDType[DTypeT] | _HasNumPyDType[DTypeT]


# A subset of `npt.DTypeLike` that can be parametrized w.r.t. `np.generic`
type _DTypeLike[ScalarT: np.generic] = (
    type[ScalarT] | np.dtype[ScalarT] | _SupportsDType[np.dtype[ScalarT]]
)


# Would create a dtype[np.void]
type _VoidDTypeLike = (
    # If a tuple, then it can be either:
    # - (flexible_dtype, itemsize)
    # - (fixed_dtype, shape)
    # - (base_dtype, new_dtype)
    # But because `_DTypeLikeNested = Any`, the first two cases are redundant

    # tuple[_DTypeLikeNested, int] | tuple[_DTypeLikeNested, _ShapeLike] |
    tuple[_DTypeLikeNested, _DTypeLikeNested]

    # [(field_name, field_dtype, field_shape), ...]
    # The type here is quite broad because NumPy accepts quite a wide
    # range of inputs inside the list; see the tests for some examples.
    | list[Any]

    # {'names': ..., 'formats': ..., 'offsets': ..., 'titles': ..., 'itemsize': ...}
    | _DTypeDict
)

# Aliases for commonly used dtype-like objects.
# Note that the precision of `np.number` subclasses is ignored herein.
type _DTypeLikeBool = type[bool] | _DTypeLike[np.bool] | _BoolCodes
type _DTypeLikeInt = type[int] | _DTypeLike[np.signedinteger] | _SignedIntegerCodes
type _DTypeLikeUInt = _DTypeLike[np.unsignedinteger] | _UnsignedIntegerCodes
type _DTypeLikeFloat = type[float] | _DTypeLike[np.floating] | _FloatingCodes
type _DTypeLikeComplex = (
    type[complex] | _DTypeLike[np.complexfloating] | _ComplexFloatingCodes
)
type _DTypeLikeComplex_co = (
    type[complex] | _DTypeLike[np.bool | np.number] | _BoolCodes | _NumberCodes
)
type _DTypeLikeDT64 = _DTypeLike[np.timedelta64] | _TD64Codes
type _DTypeLikeTD64 = _DTypeLike[np.datetime64] | _DT64Codes
type _DTypeLikeBytes = type[bytes] | _DTypeLike[np.bytes_] | _BytesCodes
type _DTypeLikeStr = type[str] | _DTypeLike[np.str_] | _StrCodes
type _DTypeLikeVoid = (
    type[memoryview] | _DTypeLike[np.void] | _VoidDTypeLike | _VoidCodes
)
type _DTypeLikeObject = type[object] | _DTypeLike[np.object_] | _ObjectCodes


# Anything that can be coerced into numpy.dtype.
# Reference: https://docs.scipy.org/doc/numpy/reference/arrays.dtypes.html
type DTypeLike = type | str | np.dtype | _SupportsDType[np.dtype] | _VoidDTypeLike

# NOTE: while it is possible to provide the dtype as a dict of
# dtype-like objects (e.g. `{'field1': ..., 'field2': ..., ...}`),
# this syntax is officially discouraged and
# therefore not included in the type-union defining `DTypeLike`.
#
# See https://github.com/numpy/numpy/issues/16891 for more details.



from collections.abc import Buffer, Callable, Collection
from typing import TYPE_CHECKING, Any, Protocol, runtime_checkable

import numpy as np

if TYPE_CHECKING:
    from numpy.dtypes import StringDType
else:
    from numpy._core.multiarray import StringDType

from ._nbit_base import _32Bit, _64Bit
from ._nested_sequence import _NestedSequence
from ._shape import _AnyShape

type NDArray[ScalarT: np.generic] = np.ndarray[_AnyShape, np.dtype[ScalarT]]

# The `_SupportsArray` protocol only cares about the default dtype
# (i.e. `dtype=None` or no `dtype` parameter at all) of the to-be returned
# array.
# Concrete implementations of the protocol are responsible for adding
# any and all remaining overloads
@runtime_checkable
class _SupportsArray[DTypeT: np.dtype](Protocol):
    def __array__(self) -> np.ndarray[Any, DTypeT]: ...


@runtime_checkable
class _SupportsArrayFunc(Protocol):
    """A protocol class representing `~class.__array_function__`."""
    def __array_function__(
        self,
        func: Callable[..., Any],
        types: Collection[type[Any]],
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> object: ...


# A subset of `npt.ArrayLike` that can be parametrized w.r.t. `np.generic`
type _ArrayLike[ScalarT: np.generic] = (
    _SupportsArray[np.dtype[ScalarT]]
    | _NestedSequence[_SupportsArray[np.dtype[ScalarT]]]
)

# A union representing array-like objects; consists of two typevars:
# One representing types that can be parametrized w.r.t. `np.dtype`
# and another one for the rest
type _DualArrayLike[DTypeT: np.dtype, BuiltinT] = (
    _SupportsArray[DTypeT]
    | _NestedSequence[_SupportsArray[DTypeT]]
    | BuiltinT
    | _NestedSequence[BuiltinT]
)

type ArrayLike = Buffer | _DualArrayLike[np.dtype, complex | bytes | str]

# `ArrayLike<X>_co`: array-like objects that can be coerced into `X`
# given the casting rules `same_kind`
type _ArrayLikeBool_co = _DualArrayLike[np.dtype[np.bool], bool]
type _ArrayLikeUInt_co = _DualArrayLike[np.dtype[np.bool | np.unsignedinteger], bool]
type _ArrayLikeInt_co = _DualArrayLike[np.dtype[np.bool | np.integer], int]
type _ArrayLikeFloat_co = _DualArrayLike[
    np.dtype[np.bool | np.integer | np.floating],
    float,
]
type _ArrayLikeComplex_co = _DualArrayLike[np.dtype[np.bool | np.number], complex]
type _ArrayLikeNumber_co = _ArrayLikeComplex_co
type _ArrayLikeTD64_co = _DualArrayLike[
    np.dtype[np.bool | np.integer | np.timedelta64],
    int,
]
type _ArrayLikeDT64_co = _ArrayLike[np.datetime64]
type _ArrayLikeObject_co = _ArrayLike[np.object_]

type _ArrayLikeVoid_co = _ArrayLike[np.void]
type _ArrayLikeBytes_co = _DualArrayLike[np.dtype[np.bytes_], bytes]
type _ArrayLikeStr_co = _DualArrayLike[np.dtype[np.str_], str]
type _ArrayLikeString_co = _DualArrayLike[StringDType, str]
type _ArrayLikeAnyString_co = _DualArrayLike[
    np.dtype[np.character] | StringDType,
    bytes | str,
]

type __Float64_co = np.floating[_64Bit] | np.float32 | np.float16 | np.integer | np.bool
type __Complex128_co = (
    np.number[_64Bit] | np.number[_32Bit] | np.float16 | np.integer | np.bool
)
type _ArrayLikeFloat64_co = _DualArrayLike[np.dtype[__Float64_co], float]
type _ArrayLikeComplex128_co = _DualArrayLike[np.dtype[__Complex128_co], complex]

# NOTE: This includes `builtins.bool`, but not `numpy.bool`.
type _ArrayLikeInt = _DualArrayLike[np.dtype[np.integer], int]



from typing import TypedDict, type_check_only

from numpy._typing import _ArrayLikeInt_co
from numpy.random.bit_generator import BitGenerator, SeedSequence

__all__ = ["PCG64"]

@type_check_only
class _PCG64Internal(TypedDict):
    state: int
    inc: int

@type_check_only
class _PCG64State(TypedDict):
    bit_generator: str
    state: _PCG64Internal
    has_uint32: int
    uinteger: int

class PCG64(BitGenerator):
    def __init__(self, seed: _ArrayLikeInt_co | SeedSequence | None = ...) -> None: ...
    def jumped(self, jumps: int = 1) -> PCG64: ...
    @property  # type: ignore[override]
    def state(self) -> _PCG64State: ...  # pyrefly: ignore[bad-override]
    @state.setter
    def state(self, value: _PCG64State) -> None: ...
    def advance(self, delta: int) -> PCG64: ...

class PCG64DXSM(BitGenerator):
    def __init__(self, seed: _ArrayLikeInt_co | SeedSequence | None = ...) -> None: ...
    def jumped(self, jumps: int = 1) -> PCG64DXSM: ...
    @property  # type: ignore[override]
    def state(self) -> _PCG64State: ...  # pyrefly: ignore[bad-override]
    @state.setter
    def state(self, value: _PCG64State) -> None: ...
    def advance(self, delta: int) -> PCG64DXSM: ...



from itertools import product, permutations

import numpy as np
import pytest
from numpy.testing import assert_allclose
from pytest import raises as assert_raises

from scipy.linalg import orthogonal_procrustes
from scipy.sparse._sputils import matrix
from scipy._lib._array_api import make_xp_test_case, xp_assert_close
from scipy.conftest import skip_xp_invalid_arg

def _centered(A, xp):
    mu = xp.mean(A, axis=0)
    return A - mu, mu

@make_xp_test_case(orthogonal_procrustes)
class TestOrthogonalProcrustes:
    def test_orthogonal_procrustes_ndim_too_small(self, xp):
        rng = np.random.RandomState(1234)
        A = xp.asarray(rng.randn(3))
        B = xp.asarray(rng.randn(3))
        assert_raises(ValueError, orthogonal_procrustes, A, B)

    def test_orthogonal_procrustes_shape_mismatch(self, xp):
        rng = np.random.RandomState(1234)
        shapes = ((3, 3), (3, 4), (4, 3), (4, 4))
        for a, b in permutations(shapes, 2):
            A = xp.asarray(rng.randn(*a))
            B = xp.asarray(rng.randn(*b))
            assert_raises(ValueError, orthogonal_procrustes, A, B)

    def test_orthogonal_procrustes_checkfinite_exception(self, xp):
        rng = np.random.RandomState(1234)
        m, n = 2, 3
        A_good = rng.randn(m, n)
        B_good = rng.randn(m, n)
        for bad_value in np.inf, -np.inf, np.nan:
            A_bad = A_good.copy()
            A_bad[1, 2] = bad_value
            B_bad = B_good.copy()
            B_bad[1, 2] = bad_value
            for A, B in ((A_good, B_bad), (A_bad, B_good), (A_bad, B_bad)):
                assert_raises(ValueError, orthogonal_procrustes, xp.asarray(A),
                              xp.asarray(B))

    def test_orthogonal_procrustes_scale_invariance(self, xp):
        rng = np.random.RandomState(1234)
        m, n = 4, 3
        for i in range(3):
            A_orig = xp.asarray(rng.randn(m, n))
            B_orig = xp.asarray(rng.randn(m, n))
            R_orig, s = orthogonal_procrustes(A_orig, B_orig)
            for A_scale in np.square(rng.randn(3)):
                for B_scale in np.square(rng.randn(3)):
                    R, s = orthogonal_procrustes(A_orig * xp.asarray(A_scale),
                                                 B_orig * xp.asarray(B_scale))
                    xp_assert_close(R, R_orig)

    @skip_xp_invalid_arg()
    def test_orthogonal_procrustes_array_conversion(self):
        rng = np.random.RandomState(1234)
        for m, n in ((6, 4), (4, 4), (4, 6)):
            A_arr = rng.randn(m, n)
            B_arr = rng.randn(m, n)
            As = (A_arr, A_arr.tolist(), matrix(A_arr))
            Bs = (B_arr, B_arr.tolist(), matrix(B_arr))
            R_arr, s = orthogonal_procrustes(A_arr, B_arr)
            AR_arr = A_arr.dot(R_arr)
            for A, B in product(As, Bs):
                R, s = orthogonal_procrustes(A, B)
                AR = A_arr.dot(R)
                assert_allclose(AR, AR_arr)

    def test_orthogonal_procrustes(self, xp):
        rng = np.random.RandomState(1234)
        for m, n in ((6, 4), (4, 4), (4, 6)):
            # Sample a random target matrix.
            B = xp.asarray(rng.randn(m, n))
            # Sample a random orthogonal matrix
            # by computing eigh of a sampled symmetric matrix.
            X = xp.asarray(rng.randn(n, n))
            w, V = xp.linalg.eigh(X.T + X)
            xp_assert_close(xp.linalg.inv(V), V.T)
            # Compute a matrix with a known orthogonal transformation that gives B.
            A = B @ V.T
            # Check that an orthogonal transformation from A to B can be recovered.
            R, s = orthogonal_procrustes(A, B)
            xp_assert_close(xp.linalg.inv(R), R.T)
            xp_assert_close(A @ R, B)
            # Create a perturbed input matrix.
            A_perturbed = A + 1e-2 * xp.asarray(rng.randn(m, n))
            # Check that the orthogonal procrustes function can find an orthogonal
            # transformation that is better than the orthogonal transformation
            # computed from the original input matrix.
            R_prime, s = orthogonal_procrustes(A_perturbed, B)
            xp_assert_close(xp.linalg.inv(R_prime), R_prime.T)
            # Compute the naive and optimal transformations of the perturbed input.
            naive_approx = A_perturbed @ R
            optim_approx = A_perturbed @ R_prime
            # Compute the Frobenius norm errors of the matrix approximations.
            naive_approx_error = xp.linalg.matrix_norm(naive_approx - B, ord='fro')
            optim_approx_error = xp.linalg.matrix_norm(optim_approx - B, ord='fro')
            # Check that the orthogonal Procrustes approximation is better.
            assert xp.all(optim_approx_error < naive_approx_error)

    def test_orthogonal_procrustes_exact_example(self, xp):
        # Check a small application.
        # It uses translation, scaling, reflection, and rotation.
        #
        #         |
        #   a  b  |
        #         |
        #   d  c  |        w
        #         |
        # --------+--- x ----- z ---
        #         |
        #         |        y
        #         |
        #
        A_orig = xp.asarray([[-3, 3], [-2, 3], [-2, 2], [-3, 2]], dtype=xp.float64)
        B_orig = xp.asarray([[3, 2], [1, 0], [3, -2], [5, 0]], dtype=xp.float64)
        A, A_mu = _centered(A_orig, xp)
        B, B_mu = _centered(B_orig, xp)
        R, s = orthogonal_procrustes(A, B)
        scale = s / xp.linalg.matrix_norm(A)**2
        B_approx = scale * A @ R + B_mu
        xp_assert_close(B_approx, B_orig, atol=1e-8)

    def test_orthogonal_procrustes_stretched_example(self, xp):
        # Try again with a target with a stretched y axis.
        A_orig = xp.asarray([[-3, 3], [-2, 3], [-2, 2], [-3, 2]], dtype=xp.float64)
        B_orig = xp.asarray([[3, 40], [1, 0], [3, -40], [5, 0]], dtype=xp.float64)
        A, A_mu = _centered(A_orig, xp)
        B, B_mu = _centered(B_orig, xp)
        R, s = orthogonal_procrustes(A, B)
        scale = s / xp.linalg.matrix_norm(A)**2
        B_approx = scale * A @ R + B_mu
        expected = xp.asarray([[3, 21], [-18, 0], [3, -21], [24, 0]], dtype=xp.float64)
        xp_assert_close(B_approx, expected, atol=1e-8)
        # Check disparity symmetry.
        expected_disparity = xp.asarray(0.4501246882793018, dtype=xp.float64)[()]
        AB_disparity = (xp.linalg.matrix_norm(B_approx - B_orig)
                        / xp.linalg.matrix_norm(B))**2
        xp_assert_close(AB_disparity, expected_disparity)
        R, s = orthogonal_procrustes(B, A)
        scale = s / xp.linalg.matrix_norm(B)**2
        A_approx = scale * B @ R + A_mu
        BA_disparity = (xp.linalg.matrix_norm(A_approx - A_orig)
                        / xp.linalg.matrix_norm(A))**2
        xp_assert_close(BA_disparity, expected_disparity)

    def test_orthogonal_procrustes_skbio_example(self, xp):
        # This transformation is also exact.
        # It uses translation, scaling, and reflection.
        #
        #   |
        #   | a
        #   | b
        #   | c d
        # --+---------
        #   |
        #   |       w
        #   |
        #   |       x
        #   |
        #   |   z   y
        #   |
        #
        A_orig = xp.asarray([[4, -2], [4, -4], [4, -6], [2, -6]], dtype=xp.float64)
        B_orig = xp.asarray([[1, 3], [1, 2], [1, 1], [2, 1]], dtype=xp.float64)
        B_standardized = xp.asarray([[-0.13363062, 0.6681531],
                                     [-0.13363062, 0.13363062],
                                     [-0.13363062, -0.40089186],
                                     [0.40089186, -0.40089186]], dtype=xp.float64)
        A, A_mu = _centered(A_orig, xp)
        B, B_mu = _centered(B_orig, xp)
        R, s = orthogonal_procrustes(A, B)
        scale = s / xp.linalg.matrix_norm(A)**2
        B_approx = scale * A @ R + B_mu
        xp_assert_close(B_approx, B_orig)
        xp_assert_close(B / xp.linalg.matrix_norm(B), B_standardized)

    def test_empty(self, xp):
        a = xp.empty((0, 0))
        r, s = orthogonal_procrustes(a, a)
        xp_assert_close(r, xp.empty((0, 0)))

        a = xp.empty((0, 3))
        r, s = orthogonal_procrustes(a, a)
        xp_assert_close(r, xp.eye(3))

    @pytest.mark.parametrize('shape', [(4, 5), (5, 5), (5, 4)])
    def test_unitary(self, shape, xp):
        # gh-12071 added support for unitary matrices; check that it
        # works as intended.
        m, n = shape
        rng = np.random.default_rng(589234981235)
        A = xp.asarray(rng.random(shape) + rng.random(shape) * 1j)
        Q = xp.asarray(rng.random((n, n)) + rng.random((n, n)) * 1j)
        Q, _ = xp.linalg.qr(Q)
        B = A @ Q
        R, scale = orthogonal_procrustes(A, B)
        xp_assert_close(R @ xp.conj(R).T, xp.eye(n, dtype=xp.complex128), atol=1e-14)
        xp_assert_close(A @ Q, B)
        if shape != (4, 5):  # solution is unique
            xp_assert_close(R, Q)
        _, s, _ = xp.linalg.svd(xp.conj(A).T @ B)
        xp_assert_close(scale, xp.sum(s))



from importlib import import_module
from pkgutil import walk_packages
import sys
import warnings

import pytest

import matplotlib
from matplotlib.testing import is_ci_environment, subprocess_run_helper

# Get the names of all matplotlib submodules,
# except for the unit tests and private modules.
module_names = []
backend_module_names = []
for m in walk_packages(path=matplotlib.__path__, prefix=f'{matplotlib.__name__}.'):
    if m.name.startswith(__package__):
        continue
    if any(x.startswith('_') for x in m.name.split('.')):
        continue
    if 'backends.backend_' in m.name:
        backend_module_names.append(m.name)
    else:
        module_names.append(m.name)


def _test_getattr(module_name, use_pytest=True):
    """
    Test that __getattr__ methods raise AttributeError for unknown keys.
    See #20822, #20855.
    """
    try:
        module = import_module(module_name)
    except (ImportError, RuntimeError, OSError) as e:
        # Skip modules that cannot be imported due to missing dependencies
        if use_pytest:
            pytest.skip(f'Cannot import {module_name} due to {e}')
        else:
            print(f'SKIP: Cannot import {module_name} due to {e}')
            return

    key = 'THIS_SYMBOL_SHOULD_NOT_EXIST'
    if hasattr(module, key):
        delattr(module, key)


@pytest.mark.parametrize('module_name', module_names)
@pytest.mark.filterwarnings('ignore::DeprecationWarning')
@pytest.mark.filterwarnings('ignore::ImportWarning')
def test_getattr(module_name):
    _test_getattr(module_name)


def _test_module_getattr():
    warnings.filterwarnings('ignore', category=DeprecationWarning)
    warnings.filterwarnings('ignore', category=ImportWarning)
    module_name = sys.argv[1]
    _test_getattr(module_name, use_pytest=False)


@pytest.mark.parametrize('module_name', backend_module_names)
def test_backend_getattr(module_name):
    proc = subprocess_run_helper(_test_module_getattr, module_name,
                                 timeout=120 if is_ci_environment() else 20)
    if 'SKIP: ' in proc.stdout:
        pytest.skip(proc.stdout.removeprefix('SKIP: '))
    print(proc.stdout)



from collections import UserDict, UserList

__all__ = ["LazyDict", "LazyList"]


class LazyDict(UserDict):
    def __init__(self, data):
        super().__init__()
        self.data = data

    def __getitem__(self, k):
        v = self.data[k]
        if callable(v):
            v = v(k)
            self.data[k] = v
        return v


class LazyList(UserList):
    def __getitem__(self, k):
        if isinstance(k, slice):
            indices = range(*k.indices(len(self)))
            return [self[i] for i in indices]
        v = self.data[k]
        if callable(v):
            v = v(k)
            self.data[k] = v
        return v

    def __add__(self, other):
        if isinstance(other, LazyList):
            other = list(other)
        elif isinstance(other, list):
            pass
        else:
            return NotImplemented
        return list(self) + other

    def __radd__(self, other):
        if not isinstance(other, list):
            return NotImplemented
        return other + list(self)



__all__ = [
    "__bibtex__",
    "__version__",
    "__version_info__",
    "set_loglevel",
    "ExecutableNotFoundError",
    "get_configdir",
    "get_cachedir",
    "get_data_path",
    "matplotlib_fname",
    "MatplotlibDeprecationWarning",
    "RcParams",
    "rc_params",
    "rc_params_from_file",
    "rcParamsDefault",
    "rcParams",
    "rcParamsOrig",
    "defaultParams",
    "rc",
    "rcdefaults",
    "rc_file_defaults",
    "rc_file",
    "rc_context",
    "use",
    "get_backend",
    "interactive",
    "is_interactive",
    "colormaps",
    "multivar_colormaps",
    "bivar_colormaps",
    "color_sequences",
]

import os
from pathlib import Path

from collections.abc import Callable, Generator
import contextlib
from packaging.version import Version

from matplotlib._api import MatplotlibDeprecationWarning
from matplotlib.typing import RcKeyType, RcGroupKeyType
from typing import Any, Literal, NamedTuple, overload
from matplotlib.typing import LogLevel


class _VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: str
    serial: int

__bibtex__: str
__version__: str
__version_info__: _VersionInfo

def set_loglevel(level: LogLevel) -> None: ...

class _ExecInfo(NamedTuple):
    executable: str
    raw_version: str
    version: Version

class ExecutableNotFoundError(FileNotFoundError): ...

def _get_executable_info(name: str) -> _ExecInfo: ...
def get_configdir() -> str: ...
def get_cachedir() -> str: ...
def get_data_path() -> str: ...
def matplotlib_fname() -> str: ...

class RcParams(dict[RcKeyType, Any]):
    validate: dict[str, Callable]
    def __init__(self, *args, **kwargs) -> None: ...
    def _set(self, key: RcKeyType, val: Any) -> None: ...
    def _get(self, key: RcKeyType) -> Any: ...

    def _update_raw(self, other_params: dict | RcParams) -> None: ...

    def _ensure_has_backend(self) -> None: ...
    def __setitem__(self, key: RcKeyType, val: Any) -> None: ...
    def __getitem__(self, key: RcKeyType) -> Any: ...
    def __iter__(self) -> Generator[RcKeyType, None, None]: ...
    def __len__(self) -> int: ...
    def find_all(self, pattern: str) -> RcParams: ...
    def copy(self) -> RcParams: ...

def rc_params(fail_on_error: bool = ...) -> RcParams: ...
def rc_params_from_file(
    fname: str | Path | os.PathLike,
    fail_on_error: bool = ...,
    use_default_template: bool = ...,
) -> RcParams: ...

rcParamsDefault: RcParams
rcParams: RcParams
rcParamsOrig: RcParams
defaultParams: dict[RcKeyType, Any]

def rc(group: RcGroupKeyType, **kwargs) -> None: ...
def rcdefaults() -> None: ...
def rc_file_defaults() -> None: ...
def rc_file(
    fname: str | Path | os.PathLike, *, use_default_template: bool = ...
) -> None: ...
@contextlib.contextmanager
def rc_context(
    rc: dict[RcKeyType, Any] | None = ..., fname: str | Path | os.PathLike | None = ...
) -> Generator[None, None, None]: ...
def use(backend: str, *, force: bool = ...) -> None: ...
@overload
def get_backend(*, auto_select: Literal[True] = True) -> str: ...
@overload
def get_backend(*, auto_select: Literal[False]) -> str | None: ...
def interactive(b: bool) -> None: ...
def is_interactive() -> bool: ...

def _preprocess_data(
    func: Callable | None = ...,
    *,
    replace_names: list[str] | None = ...,
    label_namer: str | None = ...
) -> Callable: ...

from matplotlib.cm import _colormaps as colormaps  # noqa: E402
from matplotlib.cm import _multivar_colormaps as multivar_colormaps  # noqa: E402
from matplotlib.cm import _bivar_colormaps as bivar_colormaps  # noqa: E402
from matplotlib.colors import _color_sequences as color_sequences  # noqa: E402



import argparse
import sys
from pathlib import Path

from .lib._utils_impl import get_include
from .version import __version__


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--version",
        action="version",
        version=__version__,
        help="Print the version and exit.",
    )
    parser.add_argument(
        "--cflags",
        action="store_true",
        help="Compile flag needed when using the NumPy headers.",
    )
    parser.add_argument(
        "--pkgconfigdir",
        action="store_true",
        help=("Print the pkgconfig directory in which `numpy.pc` is stored "
              "(useful for setting $PKG_CONFIG_PATH)."),
    )
    args = parser.parse_args()
    if not sys.argv[1:]:
        parser.print_help()
    if args.cflags:
        print("-I" + get_include())
    if args.pkgconfigdir:
        _path = Path(get_include()) / '..' / 'lib' / 'pkgconfig'
        print(_path.resolve())


if __name__ == "__main__":
    main()



from typing import Final

__all__ = ["NumpyVersion"]

class NumpyVersion:
    __module__ = "numpy.lib"

    vstring: Final[str]
    version: Final[str]
    major: Final[int]
    minor: Final[int]
    bugfix: Final[int]
    pre_release: Final[str]
    is_devversion: Final[bool]

    def __init__(self, /, vstring: str) -> None: ...
    def __lt__(self, other: str | NumpyVersion, /) -> bool: ...
    def __le__(self, other: str | NumpyVersion, /) -> bool: ...
    def __eq__(self, other: str | NumpyVersion, /) -> bool: ...  # type: ignore[override]
    def __ne__(self, other: str | NumpyVersion, /) -> bool: ...  # type: ignore[override]
    def __gt__(self, other: str | NumpyVersion, /) -> bool: ...
    def __ge__(self, other: str | NumpyVersion, /) -> bool: ...



import functools
import types
from scipy._lib._array_api import (
    is_cupy, is_jax, scipy_namespace_for, SCIPY_ARRAY_API, xp_capabilities
)

from ._signal_api import *   # noqa: F403
from . import _signal_api
from . import _delegators
__all__ = _signal_api.__all__


MODULE_NAME = 'signal'

# jax.scipy.signal has only partial coverage of scipy.signal, so we keep the list
# of functions we can delegate to JAX
# https://jax.readthedocs.io/en/latest/jax.scipy.html
JAX_SIGNAL_FUNCS = [
    'fftconvolve', 'convolve', 'convolve2d', 'correlate', 'correlate2d',
    'csd', 'detrend', 'istft', 'welch'
]

# some cupyx.scipy.signal functions are incompatible with their scipy counterparts
CUPY_BLACKLIST = [
    'abcd_normalize', 'bessel', 'besselap', 'envelope', 'get_window', 'lfilter_zi',
    'sosfilt_zi', 'remez',
]

# freqz_sos is a sosfreqz rename, and cupy does not have the new name yet (in v13.x)
CUPY_RENAMES = {'freqz_sos': 'sosfreqz'}


def delegate_xp(delegator, module_name):
    def inner(func):
        @functools.wraps(func)
        def wrapper(*args, **kwds):
            try:
                xp = delegator(*args, **kwds)
            except TypeError:
                # object arrays
                if func.__name__ == "tf2ss":
                    import numpy as np
                    xp = np
                else:
                    raise

            # try delegating to a cupyx/jax namesake
            if is_cupy(xp) and func.__name__ not in CUPY_BLACKLIST:
                func_name = CUPY_RENAMES.get(func.__name__, func.__name__)

                # https://github.com/cupy/cupy/issues/8336
                import importlib
                cupyx_module = importlib.import_module(f"cupyx.scipy.{module_name}")
                cupyx_func = getattr(cupyx_module, func_name)
                kwds.pop('xp', None)
                return cupyx_func(*args, **kwds)
            elif is_jax(xp) and func.__name__ in JAX_SIGNAL_FUNCS:
                spx = scipy_namespace_for(xp)
                jax_module = getattr(spx, module_name)
                jax_func = getattr(jax_module, func.__name__)
                kwds.pop('xp', None)
                return jax_func(*args, **kwds)
            else:
                # the original function
                return func(*args, **kwds)
        return wrapper
    return inner


# Although most of these functions currently exist in CuPy and some in JAX,
# there are no alternative backend tests for any of them in the current
# test suite. Each will be documented as np_only until tests are added.
untested = {
    "argrelextrema",
    "argrelmax",
    "argrelmin",
    "band_stop_obj",
    "bode",
    "check_NOLA",
    "chirp",
    "coherence",
    "csd",
    "czt",
    "czt_points",
    "dbode",
    "dfreqresp",
    "dlsim",
    "dstep",
    "find_peaks",
    "find_peaks_cwt",
    "freqresp",
    "gausspulse",
    "iirdesign", # There's no reason this shouldn't work. It just needs tests.
    "istft",
    "lombscargle",
    "lsim",
    "max_len_seq",
    "peak_prominences",
    "peak_widths",
    "periodogram",
    "place_poles",
    "sawtooth",
    "sepfir2d",
    "ss2tf",
    "ss2zpk",
    "step",
    "sweep_poly",
    "symiirorder1",
    "symiirorder2",
    "tf2ss",
    "unit_impulse",
    "welch",
    "zoom_fft",
    "zpk2ss",
}


def get_default_capabilities(func_name, delegator):
    if delegator is None or func_name in untested:
        return xp_capabilities(np_only=True)
    return xp_capabilities()

bilinear_extra_note = \
    """CuPy does not accept complex inputs.

    """

uses_choose_conv_extra_note = \
    """CuPy does not support inputs with ``ndim>1`` when ``method="auto"``
    but does support higher dimensional arrays for ``method="direct"``
    and ``method="fft"``.

    """

resample_poly_extra_note = \
    """CuPy only supports ``padtype="constant"``.

    """

upfirdn_extra_note = \
    """CuPy only supports ``mode="constant"`` and ``cval=0.0``.

    """

xord_extra_note = \
    """The ``torch`` backend on GPU does not support the case where
    `wp` and `ws` specify a Bandstop filter.

    """

convolve2d_extra_note = \
    """The JAX backend only supports ``boundary="fill"`` and ``fillvalue=0``.

    """

zpk2tf_extra_note = \
    """The CuPy and JAX backends both support only 1d input.

    """

abcd_normalize_extra_note = \
    """The result dtype when all array inputs are of integer dtype is the
    backend's current default floating point dtype.

    """

capabilities_overrides = {
    "abcd_normalize": xp_capabilities(extra_note=abcd_normalize_extra_note),
    "bessel": xp_capabilities(cpu_only=True, jax_jit=False, allow_dask_compute=True),
    "bilinear": xp_capabilities(cpu_only=True, exceptions=["cupy"],
                                jax_jit=False, allow_dask_compute=True,
                                reason="Uses np.polynomial.Polynomial",
                                extra_note=bilinear_extra_note),
    "bilinear_zpk": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                                    jax_jit=False, allow_dask_compute=True),
    "butter": xp_capabilities(cpu_only=True, exceptions=["cupy"], jax_jit=False,
                              allow_dask_compute=True),
    "buttord": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                               jax_jit=False, allow_dask_compute=True,
                               extra_note=xord_extra_note),
    "cheb1ord": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                                jax_jit=False, allow_dask_compute=True,
                                extra_note=xord_extra_note),
    "cheb2ord": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                                jax_jit=False, allow_dask_compute=True,
                                extra_note=xord_extra_note),
    "cheby1": xp_capabilities(cpu_only=True, exceptions=["cupy"], jax_jit=False,
                              allow_dask_compute=True),

    "cheby2": xp_capabilities(cpu_only=True, exceptions=["cupy"], jax_jit=False,
                              allow_dask_compute=True),
    "cont2discrete": xp_capabilities(np_only=True, exceptions=["cupy"]),
    "convolve": xp_capabilities(cpu_only=True, exceptions=["cupy", "jax.numpy"],
                                allow_dask_compute=True,
                                extra_note=uses_choose_conv_extra_note),
    "convolve2d": xp_capabilities(cpu_only=True, exceptions=["cupy", "jax.numpy"],
                                  allow_dask_compute=True,
                                  extra_note=convolve2d_extra_note),
    "correlate": xp_capabilities(cpu_only=True, exceptions=["cupy", "jax.numpy"],
                                 allow_dask_compute=True,
                                 extra_note=uses_choose_conv_extra_note),
    "correlate2d": xp_capabilities(cpu_only=True, exceptions=["cupy", "jax.numpy"],
                                   allow_dask_compute=True,
                                   extra_note=convolve2d_extra_note),
    "correlation_lags": xp_capabilities(out_of_scope=True),
    "cspline1d": xp_capabilities(cpu_only=True, exceptions=["cupy"],
                                 jax_jit=False, allow_dask_compute=True),
    "cspline1d_eval": xp_capabilities(cpu_only=True, exceptions=["cupy"],
                                      jax_jit=False, allow_dask_compute=True),
    "cspline2d": xp_capabilities(cpu_only=True, exceptions=["cupy"],
                                 jax_jit=False, allow_dask_compute=True),
    "deconvolve": xp_capabilities(cpu_only=True, exceptions=["cupy"],
                                  jax_jit=False, allow_dask_compute=True),
    "decimate": xp_capabilities(np_only=True, exceptions=["cupy"]),
    "detrend": xp_capabilities(cpu_only=True, exceptions=["cupy", "jax.numpy"],
                               allow_dask_compute=True),
    "dimpulse": xp_capabilities(np_only=True, exceptions=["cupy"]),
    "dlti": xp_capabilities(np_only=True,
                            reason="works in CuPy but delegation isn't set up yet"),
    "ellip": xp_capabilities(cpu_only=True, exceptions=["cupy"], jax_jit=False,
                             allow_dask_compute=True,
                             reason="scipy.special.ellipk"),
    "ellipord": xp_capabilities(cpu_only=True, exceptions=["cupy"],
                                jax_jit=False, allow_dask_compute=True,
                                reason="scipy.special.ellipk"),
    "findfreqs": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                                 jax_jit=False, allow_dask_compute=True),
    "firls": xp_capabilities(cpu_only=True, allow_dask_compute=True, jax_jit=False,
                             reason="lstsq"),
    "firwin": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                              jax_jit=False, allow_dask_compute=True),
    "firwin2": xp_capabilities(cpu_only=True, exceptions=["cupy"],
                               jax_jit=False, allow_dask_compute=True,
                               reason="firwin uses np.interp"),
    "fftconvolve": xp_capabilities(cpu_only=True,
                                   exceptions=["cupy", "jax.numpy", "torch"]),
    "freqs": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                             jax_jit=False, allow_dask_compute=True),
    "freqs_zpk": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                                 jax_jit=False, allow_dask_compute=True),
    "freqz": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                             jax_jit=False, allow_dask_compute=True),
    "freqz_sos": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                                 jax_jit=False, allow_dask_compute=True),
    "group_delay": xp_capabilities(cpu_only=True, exceptions=["cupy"],
                                   jax_jit=False, allow_dask_compute=True),
    "invres": xp_capabilities(np_only=True, exceptions=["cupy"]),
    "invresz": xp_capabilities(np_only=True, exceptions=["cupy"]),
    "iircomb": xp_capabilities(xfail_backends=[("jax.numpy", "inaccurate")]),
    "iirfilter": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                                 jax_jit=False, allow_dask_compute=True),
    "kaiser_atten": xp_capabilities(
        out_of_scope=True, reason="scalars in, scalars out"
    ),
    "kaiser_beta": xp_capabilities(out_of_scope=True, reason="scalars in, scalars out"),
    "kaiserord": xp_capabilities(out_of_scope=True, reason="scalars in, scalars out"),
    "lfilter": xp_capabilities(cpu_only=True, exceptions=["cupy"],
                               allow_dask_compute=True, jax_jit=False),
    "lfilter_zi": xp_capabilities(cpu_only=True, allow_dask_compute=True,
                                  jax_jit=False),
    "lfiltic": xp_capabilities(cpu_only=True, exceptions=["cupy"],
                               allow_dask_compute=True),
    "lp2bp": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                             allow_dask_compute=True, jax_jit=False),
    "lp2bp_zpk": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                                 allow_dask_compute=True, jax_jit=False),
    "lp2bs": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                             allow_dask_compute=True, jax_jit=False),
    "lp2bs_zpk": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                                 allow_dask_compute=True, jax_jit=False),
    "lp2lp": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                             allow_dask_compute=True, jax_jit=False),
    "lp2lp_zpk": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                                 allow_dask_compute=True, jax_jit=False),
    "lp2hp": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                             allow_dask_compute=True, jax_jit=False),
    "lp2hp_zpk": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                                 allow_dask_compute=True, jax_jit=False),
    "lti": xp_capabilities(np_only=True,
                            reason="works in CuPy but delegation isn't set up yet"),
    "medfilt": xp_capabilities(cpu_only=True, exceptions=["cupy"],
                               allow_dask_compute=True, jax_jit=False,
                               reason="uses scipy.ndimage.rank_filter"),
    "medfilt2d": xp_capabilities(cpu_only=True, exceptions=["cupy"],
                                 allow_dask_compute=True, jax_jit=False,
                                 reason="c extension module"),
    "minimum_phase": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                                     allow_dask_compute=True, jax_jit=False),
    "normalize": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                                 jax_jit=False, allow_dask_compute=True),
    "oaconvolve": xp_capabilities(
        cpu_only=True, exceptions=["cupy", "torch"],
        xfail_backends=[("dask.array", "wrong answer")],
    ),
    "order_filter": xp_capabilities(cpu_only=True, exceptions=["cupy"],
                                    allow_dask_compute=True, jax_jit=False,
                                    reason="uses scipy.ndimage.rank_filter"),
    "qspline1d": xp_capabilities(cpu_only=True, exceptions=["cupy"],
                                 jax_jit=False, allow_dask_compute=True),
    "qspline1d_eval": xp_capabilities(cpu_only=True, exceptions=["cupy"],
                                      jax_jit=False, allow_dask_compute=True),
    "qspline2d": xp_capabilities(np_only=True, exceptions=["cupy"]),
    "remez": xp_capabilities(cpu_only=True, allow_dask_compute=True, jax_jit=False),
    "resample": xp_capabilities(
        cpu_only=True, exceptions=["cupy"],
        skip_backends=[
            ("dask.array", "XXX something in dask"),
            ("jax.numpy", "XXX: immutable arrays"),
        ]
    ),
    "resample_poly": xp_capabilities(
        cpu_only=True, exceptions=["cupy"],
        jax_jit=False, skip_backends=[("dask.array", "XXX something in dask")],
        extra_note=resample_poly_extra_note,
    ),
    "residue": xp_capabilities(np_only=True, exceptions=["cupy"]),
    "residuez": xp_capabilities(np_only=True, exceptions=["cupy"]),
    "savgol_filter": xp_capabilities(cpu_only=True, exceptions=["cupy"],
                                     jax_jit=False,
                                     reason="convolve1d is cpu-only"),
    "sawtooth": xp_capabilities(jax_jit=False,
                                skip_backends=[("dask.array", "dask tests fail")]),
    "sepfir2d": xp_capabilities(np_only=True),
    "sos2zpk": xp_capabilities(cpu_only=True, exceptions=["cupy"], jax_jit=False,
                               allow_dask_compute=True),
    "sos2tf": xp_capabilities(cpu_only=True, exceptions=["cupy"], jax_jit=False,
                              allow_dask_compute=True),
    "sosfilt": xp_capabilities(cpu_only=True, exceptions=["cupy"], jax_jit=False,
                               allow_dask_compute=True),
    "sosfilt_zi": xp_capabilities(cpu_only=True, allow_dask_compute=True,
                                  jax_jit=False),
    "sosfiltfilt": xp_capabilities(
        cpu_only=True, exceptions=["cupy"],
        skip_backends=[
            (
                "dask.array",
                "sosfiltfilt directly sets shape attributes on arrays"
                " which dask doesn't like"
            ),
            ("torch", "negative strides"),
            ("jax.numpy", "sosfilt works in-place"),
        ],
    ),
    "sosfreqz": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                                jax_jit=False, allow_dask_compute=True),
    "spline_filter": xp_capabilities(cpu_only=True, exceptions=["cupy"],
                                     jax_jit=False, allow_dask_compute=True),
    "tf2sos": xp_capabilities(cpu_only=True, exceptions=["cupy"], jax_jit=False,
                              allow_dask_compute=True),
    "tf2zpk": xp_capabilities(cpu_only=True, exceptions=["cupy"], jax_jit=False,
                              allow_dask_compute=True),
    "unique_roots": xp_capabilities(np_only=True, exceptions=["cupy"]),
    "upfirdn": xp_capabilities(cpu_only=True, exceptions=["cupy"], jax_jit=False,
                               allow_dask_compute=True,
                               reason="Cython implementation",
                               extra_note=upfirdn_extra_note),
    "vectorstrength": xp_capabilities(cpu_only=True, exceptions=["cupy", "torch"],
                                      allow_dask_compute=True, jax_jit=False),
    "wiener": xp_capabilities(cpu_only=True, exceptions=["cupy", "jax.numpy"],
                              allow_dask_compute=True, jax_jit=False,
                              reason="uses scipy.signal.correlate"),
    "zpk2sos": xp_capabilities(cpu_only=True, exceptions=["cupy"], jax_jit=False,
                               allow_dask_compute=True),
    "zpk2tf": xp_capabilities(cpu_only=True, exceptions=["cupy"], jax_jit=False,
                              allow_dask_compute=True,
                              extra_note=zpk2tf_extra_note),
    "spectrogram": xp_capabilities(out_of_scope=True),  # legacy
    "stft": xp_capabilities(out_of_scope=True),  # legacy
    "istft": xp_capabilities(out_of_scope=True),  # legacy
    "check_COLA": xp_capabilities(out_of_scope=True),  # legacy
}


# ### decorate ###
for obj_name in _signal_api.__all__:
    bare_obj = getattr(_signal_api, obj_name)
    delegator = getattr(_delegators, obj_name + "_signature", None)

    if SCIPY_ARRAY_API and delegator is not None:
        f = delegate_xp(delegator, MODULE_NAME)(bare_obj)
    else:
        f = bare_obj

    if not isinstance(f, types.ModuleType):
        capabilities = capabilities_overrides.get(
            obj_name, get_default_capabilities(obj_name, delegator)
        )
        f = capabilities(f)  # pyrefly:ignore[not-callable]

    # add the decorated function to the namespace, to be imported in __init__.py
    vars()[obj_name] = f



from __future__ import annotations

from fontTools.pens.basePen import AbstractPen, DecomposingPen
from fontTools.pens.pointPen import (
    AbstractPointPen,
    DecomposingPointPen,
    ReverseFlipped,
)
from fontTools.pens.recordingPen import RecordingPen


class _PassThruComponentsMixin(object):
    def addComponent(self, glyphName, transformation, **kwargs):
        self._outPen.addComponent(glyphName, transformation, **kwargs)


class FilterPen(_PassThruComponentsMixin, AbstractPen):
    """Base class for pens that apply some transformation to the coordinates
    they receive and pass them to another pen.

    You can override any of its methods. The default implementation does
    nothing, but passes the commands unmodified to the other pen.

    >>> from fontTools.pens.recordingPen import RecordingPen
    >>> rec = RecordingPen()
    >>> pen = FilterPen(rec)
    >>> v = iter(rec.value)

    >>> pen.moveTo((0, 0))
    >>> next(v)
    ('moveTo', ((0, 0),))

    >>> pen.lineTo((1, 1))
    >>> next(v)
    ('lineTo', ((1, 1),))

    >>> pen.curveTo((2, 2), (3, 3), (4, 4))
    >>> next(v)
    ('curveTo', ((2, 2), (3, 3), (4, 4)))

    >>> pen.qCurveTo((5, 5), (6, 6), (7, 7), (8, 8))
    >>> next(v)
    ('qCurveTo', ((5, 5), (6, 6), (7, 7), (8, 8)))

    >>> pen.closePath()
    >>> next(v)
    ('closePath', ())

    >>> pen.moveTo((9, 9))
    >>> next(v)
    ('moveTo', ((9, 9),))

    >>> pen.endPath()
    >>> next(v)
    ('endPath', ())

    >>> pen.addComponent('foo', (1, 0, 0, 1, 0, 0))
    >>> next(v)
    ('addComponent', ('foo', (1, 0, 0, 1, 0, 0)))
    """

    def __init__(self, outPen):
        self._outPen = outPen
        self.current_pt = None

    def moveTo(self, pt):
        self._outPen.moveTo(pt)
        self.current_pt = pt

    def lineTo(self, pt):
        self._outPen.lineTo(pt)
        self.current_pt = pt

    def curveTo(self, *points):
        self._outPen.curveTo(*points)
        self.current_pt = points[-1]

    def qCurveTo(self, *points):
        self._outPen.qCurveTo(*points)
        self.current_pt = points[-1]

    def closePath(self):
        self._outPen.closePath()
        self.current_pt = None

    def endPath(self):
        self._outPen.endPath()
        self.current_pt = None


class ContourFilterPen(_PassThruComponentsMixin, RecordingPen):
    """A "buffered" filter pen that accumulates contour data, passes
    it through a ``filterContour`` method when the contour is closed or ended,
    and finally draws the result with the output pen.

    Components are passed through unchanged.
    """

    def __init__(self, outPen):
        super(ContourFilterPen, self).__init__()
        self._outPen = outPen

    def closePath(self):
        super(ContourFilterPen, self).closePath()
        self._flushContour()

    def endPath(self):
        super(ContourFilterPen, self).endPath()
        self._flushContour()

    def _flushContour(self):
        result = self.filterContour(self.value)
        if result is not None:
            self.value = result
        self.replay(self._outPen)
        self.value = []

    def filterContour(self, contour):
        """Subclasses must override this to perform the filtering.

        The contour is a list of pen (operator, operands) tuples.
        Operators are strings corresponding to the AbstractPen methods:
        "moveTo", "lineTo", "curveTo", "qCurveTo", "closePath" and
        "endPath". The operands are the positional arguments that are
        passed to each method.

        If the method doesn't return a value (i.e. returns None), it's
        assumed that the argument was modified in-place.
        Otherwise, the return value is drawn with the output pen.
        """
        return  # or return contour


class FilterPointPen(_PassThruComponentsMixin, AbstractPointPen):
    """Baseclass for point pens that apply some transformation to the
    coordinates they receive and pass them to another point pen.

    You can override any of its methods. The default implementation does
    nothing, but passes the commands unmodified to the other pen.

    >>> from fontTools.pens.recordingPen import RecordingPointPen
    >>> rec = RecordingPointPen()
    >>> pen = FilterPointPen(rec)
    >>> v = iter(rec.value)
    >>> pen.beginPath(identifier="abc")
    >>> next(v)
    ('beginPath', (), {'identifier': 'abc'})
    >>> pen.addPoint((1, 2), "line", False)
    >>> next(v)
    ('addPoint', ((1, 2), 'line', False, None), {})
    >>> pen.addComponent("a", (2, 0, 0, 2, 10, -10), identifier="0001")
    >>> next(v)
    ('addComponent', ('a', (2, 0, 0, 2, 10, -10)), {'identifier': '0001'})
    >>> pen.endPath()
    >>> next(v)
    ('endPath', (), {})
    """

    def __init__(self, outPen):
        self._outPen = outPen

    def beginPath(self, identifier=None, **kwargs):
        kwargs = dict(kwargs)
        if identifier is not None:
            kwargs["identifier"] = identifier
        self._outPen.beginPath(**kwargs)

    def endPath(self):
        self._outPen.endPath()

    def addPoint(
        self,
        pt,
        segmentType=None,
        smooth=False,
        name=None,
        identifier=None,
        **kwargs,
    ):
        kwargs = dict(kwargs)
        if identifier is not None:
            kwargs["identifier"] = identifier
        self._outPen.addPoint(pt, segmentType, smooth, name, **kwargs)


class _DecomposingFilterMixinBase:
    """Base mixin class with common `addComponent` logic for decomposing filter pens."""

    def addComponent(self, baseGlyphName, transformation, **kwargs):
        # only decompose the component if it's included in the set
        if self.include is None or baseGlyphName in self.include:
            # if we're decomposing nested components, temporarily set include to None
            include_bak = self.include
            if self.decomposeNested and self.include:
                self.include = None
            try:
                super().addComponent(baseGlyphName, transformation, **kwargs)
            finally:
                if self.include != include_bak:
                    self.include = include_bak
        else:
            _PassThruComponentsMixin.addComponent(
                self, baseGlyphName, transformation, **kwargs
            )


class _DecomposingFilterPenMixin(_DecomposingFilterMixinBase):
    """Mixin class that decomposes components as regular contours for segment pens.

    Used by DecomposingFilterPen.

    Takes two required parameters, another segment pen 'outPen' to draw
    with, and a 'glyphSet' dict of drawable glyph objects to draw components from.

    The 'skipMissingComponents' and 'reverseFlipped' optional arguments work the
    same as in the DecomposingPen. reverseFlipped is bool only (True/False).

    In addition, the decomposing filter pens also take the following two options:

    'include' is an optional set of component base glyph names to consider for
    decomposition; the default include=None means decompose all components no matter
    the base glyph name).

    'decomposeNested' (bool) controls whether to recurse decomposition into nested
    components of components (this only matters when 'include' was also provided);
    if False, only decompose top-level components included in the set, but not
    also their children.
    """

    # raises MissingComponentError if base glyph is not found in glyphSet
    skipMissingComponents = False

    def __init__(
        self,
        outPen,
        glyphSet,
        skipMissingComponents=None,
        reverseFlipped: bool = False,
        include: set[str] | None = None,
        decomposeNested: bool = True,
        **kwargs,
    ):
        assert isinstance(
            reverseFlipped, bool
        ), f"Expected bool, got {type(reverseFlipped).__name__}"
        super().__init__(
            outPen=outPen,
            glyphSet=glyphSet,
            skipMissingComponents=skipMissingComponents,
            reverseFlipped=reverseFlipped,
            **kwargs,
        )
        self.include = include
        self.decomposeNested = decomposeNested


class _DecomposingFilterPointPenMixin(_DecomposingFilterMixinBase):
    """Mixin class that decomposes components as regular contours for point pens.

    Takes two required parameters, another point pen 'outPen' to draw
    with, and a 'glyphSet' dict of drawable glyph objects to draw components from.

    The 'skipMissingComponents' and 'reverseFlipped' optional arguments work the
    same as in the DecomposingPointPen. reverseFlipped accepts bool | ReverseFlipped
    (see DecomposingPointPen).

    In addition, the decomposing filter pens also take the following two options:

    'include' is an optional set of component base glyph names to consider for
    decomposition; the default include=None means decompose all components no matter
    the base glyph name).

    'decomposeNested' (bool) controls whether to recurse decomposition into nested
    components of components (this only matters when 'include' was also provided);
    if False, only decompose top-level components included in the set, but not
    also their children.
    """

    # raises MissingComponentError if base glyph is not found in glyphSet
    skipMissingComponents = False

    def __init__(
        self,
        outPen,
        glyphSet,
        skipMissingComponents=None,
        reverseFlipped: bool | ReverseFlipped = False,
        include: set[str] | None = None,
        decomposeNested: bool = True,
        **kwargs,
    ):
        super().__init__(
            outPen=outPen,
            glyphSet=glyphSet,
            skipMissingComponents=skipMissingComponents,
            reverseFlipped=reverseFlipped,
            **kwargs,
        )
        self.include = include
        self.decomposeNested = decomposeNested


class DecomposingFilterPen(_DecomposingFilterPenMixin, DecomposingPen, FilterPen):
    """Filter pen that draws components as regular contours."""

    pass


class DecomposingFilterPointPen(
    _DecomposingFilterPointPenMixin, DecomposingPointPen, FilterPointPen
):
    """Filter point pen that draws components as regular contours."""

    pass


class ContourFilterPointPen(_PassThruComponentsMixin, AbstractPointPen):
    """A "buffered" filter point pen that accumulates contour data, passes
    it through a ``filterContour`` method when the contour is closed or ended,
    and finally draws the result with the output point pen.

    Components are passed through unchanged.

    The ``filterContour`` method can modify the contour in-place (return None)
    or return a new contour to replace it.
    """

    def __init__(self, outPen):
        self._outPen = outPen
        self.currentContour = None
        self.currentContourKwargs = None

    def beginPath(self, identifier=None, **kwargs):
        if self.currentContour is not None:
            raise ValueError("Path already begun")
        kwargs = dict(kwargs)
        if identifier is not None:
            kwargs["identifier"] = identifier
        self.currentContour = []
        self.currentContourKwargs = kwargs

    def endPath(self):
        if self.currentContour is None:
            raise ValueError("Path not begun")
        self._flushContour()
        self.currentContour = None
        self.currentContourKwargs = None

    def _flushContour(self):
        """Flush the current contour to the output pen."""
        result = self.filterContour(self.currentContour)
        if result is not None:
            self.currentContour = result

        # Draw the filtered contour
        self._outPen.beginPath(**self.currentContourKwargs)
        for pt, segmentType, smooth, name, kwargs in self.currentContour:
            self._outPen.addPoint(pt, segmentType, smooth, name, **kwargs)
        self._outPen.endPath()

    def filterContour(self, contour):
        """Subclasses must override this to perform the filtering.

        The contour is a list of (pt, segmentType, smooth, name, kwargs) tuples.
        If the method doesn't return a value (i.e. returns None), it's
        assumed that the contour was modified in-place.
        Otherwise, the return value replaces the original contour.
        """
        return  # or return contour

    def addPoint(
        self,
        pt,
        segmentType=None,
        smooth=False,
        name=None,
        identifier=None,
        **kwargs,
    ):
        if self.currentContour is None:
            raise ValueError("Path not begun")
        kwargs = dict(kwargs)
        if identifier is not None:
            kwargs["identifier"] = identifier
        self.currentContour.append((pt, segmentType, smooth, name, kwargs))


class OnCurveFirstPointPen(ContourFilterPointPen):
    """Filter point pen that ensures closed contours start with an on-curve point.

    If a closed contour starts with an off-curve point (segmentType=None), it rotates
    the points list so that the first on-curve point (segmentType != None) becomes
    the start point. Open contours and contours already starting with on-curve points
    are passed through unchanged.

    >>> from fontTools.pens.recordingPen import RecordingPointPen
    >>> rec = RecordingPointPen()
    >>> pen = OnCurveFirstPointPen(rec)
    >>> # Closed contour starting with off-curve - will be rotated
    >>> pen.beginPath()
    >>> pen.addPoint((0, 0), None)  # off-curve
    >>> pen.addPoint((100, 100), "line")  # on-curve - will become start
    >>> pen.addPoint((200, 0), None)  # off-curve
    >>> pen.addPoint((300, 100), "curve")  # on-curve
    >>> pen.endPath()
    >>> # The contour should now start with (100, 100) "line"
    >>> rec.value[0]
    ('beginPath', (), {})
    >>> rec.value[1]
    ('addPoint', ((100, 100), 'line', False, None), {})
    >>> rec.value[2]
    ('addPoint', ((200, 0), None, False, None), {})
    >>> rec.value[3]
    ('addPoint', ((300, 100), 'curve', False, None), {})
    >>> rec.value[4]
    ('addPoint', ((0, 0), None, False, None), {})
    """

    def filterContour(self, contour):
        """Rotate closed contour to start with first on-curve point if needed."""
        if not contour:
            return

        # Check if it's a closed contour (no "move" segmentType)
        is_closed = contour[0][1] != "move"

        if is_closed and contour[0][1] is None:
            # Closed contour starting with off-curve - need to rotate
            # Find the first on-curve point
            for i, (pt, segmentType, smooth, name, kwargs) in enumerate(contour):
                if segmentType is not None:
                    # Rotate the points list so it starts with the first on-curve point
                    return contour[i:] + contour[:i]



import multiprocessing

import pytest

import numpy as np
from numpy.testing import IS_WASM

pytestmark = pytest.mark.thread_unsafe(
    reason="tests in this module are explicitly multi-processed"
)

def bool_array_writer(shm_name, n):
    # writer routine for test_read_write_bool_array
    import time
    from multiprocessing import shared_memory
    shm = shared_memory.SharedMemory(name=shm_name)
    arr = np.ndarray(n, dtype=np.bool_, buffer=shm.buf)
    for i in range(n):
        arr[i] = True
        time.sleep(0.00001)

def bool_array_reader(shm_name, n):
    # reader routine for test_read_write_bool_array
    from multiprocessing import shared_memory
    shm = shared_memory.SharedMemory(name=shm_name)
    arr = np.ndarray(n, dtype=np.bool_, buffer=shm.buf)
    for i in range(n):
        while not arr[i]:
            pass

@pytest.mark.skipif(IS_WASM,
                    reason="WASM does not support _posixshmem")
@pytest.mark.slow
def test_read_write_bool_array():
    # See: gh-30389
    #
    # Prior to Python 3.13, boolean scalar singletons (np.True / np.False) were
    # regular reference-counted objects. Due to the double evaluation in
    # PyArrayScalar_RETURN_BOOL_FROM_LONG, concurrent reads and writes of a
    # boolean array could corrupt their refcounts, potentially causing a crash
    # (e.g., `free(): invalid pointer`).
    #
    # This test creates a multi-process race between a writer and a reader to
    # ensure that NumPy does not exhibit such failures.
    from concurrent.futures import ProcessPoolExecutor
    from multiprocessing import shared_memory
    n = 10000
    shm = shared_memory.SharedMemory(create=True, size=n)
    ctx = multiprocessing.get_context("spawn")
    with ProcessPoolExecutor(max_workers=2, mp_context=ctx) as executor:
        f_writer = executor.submit(bool_array_writer, shm.name, n)
        f_reader = executor.submit(bool_array_reader, shm.name, n)
    shm.unlink()
    f_writer.result()
    f_reader.result()



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from contourpy._contourpy import FillType, LineType
from contourpy.array import (
    concat_codes_or_none,
    concat_offsets_or_none,
    concat_points_or_none,
    concat_points_or_none_with_nan,
)
from contourpy.enum_util import as_fill_type, as_line_type
from contourpy.typecheck import check_filled, check_lines

if TYPE_CHECKING:
    import contourpy._contourpy as cpy


def dechunk_filled(filled: cpy.FillReturn, fill_type: FillType | str) -> cpy.FillReturn:
    """Return the specified filled contours with chunked data moved into the first chunk.

    Filled contours that are not chunked (``FillType.OuterCode`` and ``FillType.OuterOffset``) and
    those that are but only contain a single chunk are returned unmodified. Individual polygons are
    unchanged, they are not geometrically combined.

    Args:
        filled (sequence of arrays): Filled contour data, such as returned by
            :meth:`.ContourGenerator.filled`.
        fill_type (FillType or str): Type of :meth:`~.ContourGenerator.filled` as enum or string
            equivalent.

    Return:
        Filled contours in a single chunk.

    .. versionadded:: 1.2.0
    """
    fill_type = as_fill_type(fill_type)

    if fill_type in (FillType.OuterCode, FillType.OuterOffset):
        # No-op if fill_type is not chunked.
        return filled

    check_filled(filled, fill_type)
    if len(filled[0]) < 2:
        # No-op if just one chunk.
        return filled

    if TYPE_CHECKING:
        filled = cast(cpy.FillReturn_Chunk, filled)
    points = concat_points_or_none(filled[0])

    if fill_type == FillType.ChunkCombinedCode:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_ChunkCombinedCode, filled)
        if points is None:
            ret1: cpy.FillReturn_ChunkCombinedCode = ([None], [None])
        else:
            ret1 = ([points], [concat_codes_or_none(filled[1])])
        return ret1
    elif fill_type == FillType.ChunkCombinedOffset:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_ChunkCombinedOffset, filled)
        if points is None:
            ret2: cpy.FillReturn_ChunkCombinedOffset = ([None], [None])
        else:
            ret2 = ([points], [concat_offsets_or_none(filled[1])])
        return ret2
    elif fill_type == FillType.ChunkCombinedCodeOffset:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_ChunkCombinedCodeOffset, filled)
        if points is None:
            ret3: cpy.FillReturn_ChunkCombinedCodeOffset = ([None], [None], [None])
        else:
            outer_offsets = concat_offsets_or_none(filled[2])
            ret3 = ([points], [concat_codes_or_none(filled[1])], [outer_offsets])
        return ret3
    elif fill_type == FillType.ChunkCombinedOffsetOffset:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_ChunkCombinedOffsetOffset, filled)
        if points is None:
            ret4: cpy.FillReturn_ChunkCombinedOffsetOffset = ([None], [None], [None])
        else:
            outer_offsets = concat_offsets_or_none(filled[2])
            ret4 = ([points], [concat_offsets_or_none(filled[1])], [outer_offsets])
        return ret4
    else:
        raise ValueError(f"Invalid FillType {fill_type}")


def dechunk_lines(lines: cpy.LineReturn, line_type: LineType | str) -> cpy.LineReturn:
    """Return the specified contour lines with chunked data moved into the first chunk.

    Contour lines that are not chunked (``LineType.Separate`` and ``LineType.SeparateCode``) and
    those that are but only contain a single chunk are returned unmodified. Individual lines are
    unchanged, they are not geometrically combined.

    Args:
        lines (sequence of arrays): Contour line data, such as returned by
            :meth:`.ContourGenerator.lines`.
        line_type (LineType or str): Type of :meth:`~.ContourGenerator.lines` as enum or string
            equivalent.

    Return:
        Contour lines in a single chunk.

    .. versionadded:: 1.2.0
    """
    line_type = as_line_type(line_type)

    if line_type in (LineType.Separate, LineType.SeparateCode):
        # No-op if line_type is not chunked.
        return lines

    check_lines(lines, line_type)
    if len(lines[0]) < 2:
        # No-op if just one chunk.
        return lines

    if TYPE_CHECKING:
        lines = cast(cpy.LineReturn_Chunk, lines)

    if line_type == LineType.ChunkCombinedCode:
        if TYPE_CHECKING:
            lines = cast(cpy.LineReturn_ChunkCombinedCode, lines)
        points = concat_points_or_none(lines[0])
        if points is None:
            ret1: cpy.LineReturn_ChunkCombinedCode = ([None], [None])
        else:
            ret1 = ([points], [concat_codes_or_none(lines[1])])
        return ret1
    elif line_type == LineType.ChunkCombinedOffset:
        if TYPE_CHECKING:
            lines = cast(cpy.LineReturn_ChunkCombinedOffset, lines)
        points = concat_points_or_none(lines[0])
        if points is None:
            ret2: cpy.LineReturn_ChunkCombinedOffset = ([None], [None])
        else:
            ret2 = ([points], [concat_offsets_or_none(lines[1])])
        return ret2
    elif line_type == LineType.ChunkCombinedNan:
        if TYPE_CHECKING:
            lines = cast(cpy.LineReturn_ChunkCombinedNan, lines)
        points = concat_points_or_none_with_nan(lines[0])
        ret3: cpy.LineReturn_ChunkCombinedNan = ([points],)
        return ret3
    else:
        raise ValueError(f"Invalid LineType {line_type}")


def dechunk_multi_filled(
    multi_filled: list[cpy.FillReturn],
    fill_type: FillType | str,
) -> list[cpy.FillReturn]:
    """Return multiple sets of filled contours with chunked data moved into the first chunks.

    Filled contours that are not chunked (``FillType.OuterCode`` and ``FillType.OuterOffset``) and
    those that are but only contain a single chunk are returned unmodified. Individual polygons are
    unchanged, they are not geometrically combined.

    Args:
        multi_filled (nested sequence of arrays): Filled contour data, such as returned by
            :meth:`.ContourGenerator.multi_filled`.
        fill_type (FillType or str): Type of :meth:`~.ContourGenerator.filled` as enum or string
            equivalent.

    Return:
        Multiple sets of filled contours in a single chunk.

    .. versionadded:: 1.3.0
    """
    fill_type = as_fill_type(fill_type)

    if fill_type in (FillType.OuterCode, FillType.OuterOffset):
        # No-op if fill_type is not chunked.
        return multi_filled

    return [dechunk_filled(filled, fill_type) for filled in multi_filled]


def dechunk_multi_lines(
    multi_lines: list[cpy.LineReturn],
    line_type: LineType | str,
) -> list[cpy.LineReturn]:
    """Return multiple sets of contour lines with all chunked data moved into the first chunks.

    Contour lines that are not chunked (``LineType.Separate`` and ``LineType.SeparateCode``) and
    those that are but only contain a single chunk are returned unmodified. Individual lines are
    unchanged, they are not geometrically combined.

    Args:
        multi_lines (nested sequence of arrays): Contour line data, such as returned by
            :meth:`.ContourGenerator.multi_lines`.
        line_type (LineType or str): Type of :meth:`~.ContourGenerator.lines` as enum or string
            equivalent.

    Return:
        Multiple sets of contour lines in a single chunk.

    .. versionadded:: 1.3.0
    """
    line_type = as_line_type(line_type)

    if line_type in (LineType.Separate, LineType.SeparateCode):
        # No-op if line_type is not chunked.
        return multi_lines

    return [dechunk_lines(lines, line_type) for lines in multi_lines]
